#include <iostream>
#include "crypt.h"
#include "MersenneTwister.h"

// Methylation probability for progenitor cells
#define METH_PROG_INF (randomNumberGenerator.rand(100000.0) < 6.0)
#define METH_PROG_NORM (randomNumberGenerator.rand(100000.0) < 0.02)

// Methylation probability for stem cells
#define METH_STEM_INF (randomNumberGenerator.rand(100000.0) < 0.003)
#define METH_STEM_NORM false

// Life span: 3 days without infection, shorter during infection
#define NORMAL_LIFE_SPAN 4300
#define INFECTED_LIFE_SPAN 3500

// Maximum size
#define LIMIT_SIZE 1000

// Crypt size in normal conditions: about 100 diff. cells on each side
#define NORMAL_SIZE 100

MTRand randomNumberGenerator;
using namespace std;


/*********************/
/* Crypt constructor */
/*********************/

Crypt::Crypt() {
	// Initialisation of atomic variables
	cellPopulation=2*LIMIT_SIZE+7;
	isInfected=false;
	statusLength=0;
	infectionTime=0;
	methylationLevel=0;
	sizeRight=100; sizeLeft=100;
	cells = new Cell [cellPopulation];
	// Initialisation of the cell array
	for(int i=0; i<LIMIT_SIZE-sizeRight; ++i) {
		cells[i].Erase();
		cells[cellPopulation-1-i].Erase();
	}
	// Initialisation of the differentiated cells
	for(int i=LIMIT_SIZE-sizeRight; i<LIMIT_SIZE; ++i) {
		cells[i].Create(DifferentiatedCell,false,43*i/(LIMIT_SIZE/sizeLeft));
		cells[cellPopulation-1-i].Create(DifferentiatedCell,false,43*i/(LIMIT_SIZE/sizeRight));
	}
	// Initialisation of the progenitor cells
	for(int i=LIMIT_SIZE; i<LIMIT_SIZE+3; ++i) {
		cells[i].Create(ProgenitorCell,false,0);
		cells[cellPopulation-1-i].Create(ProgenitorCell,false,0);
		progenitorCells[i-LIMIT_SIZE]=i;
		progenitorCells[cellPopulation-2-i-LIMIT_SIZE]=cellPopulation-1-i;
	}
	// Initialisation of the stem cell
	cells[LIMIT_SIZE+3].Create(StemCell,false,0);
	stemCell=LIMIT_SIZE+3;

	// Initialisation of the "producetion rate" used during infection
	// (see Thesis for details)
	int randomValue = randomNumberGenerator.randInt(100);
	if(randomValue <= 30)
		infectedSizeParameter = 30000 + randomNumberGenerator.randInt(15000); // uniform in [30,000; 45,000]
	else {
		if(randomValue <= 60)
			infectedSizeParameter = 45000 + randomNumberGenerator.randInt(30000); // uniform in [45,000; 75,000]
		else {
			if(randomValue <= 80)
				infectedSizeParameter = 75000 + randomNumberGenerator.randInt(25000); // uniform in [75,000; 100,000]
			else
				infectedSizeParameter = 100000 + randomNumberGenerator.randInt(10000); // uniform in [100,000; 110,000]
		}
	}

	// Initialisation completed
}


/********************/
/* Crypt destructor */
/********************/

Crypt::~Crypt() {
	delete cells;
}


/**************************************/
/* Starting an infection in the crypt */
/**************************************/

void Crypt::StartInfection() {
	isInfected=true;
	statusLength=0;
}

/***************************************/
/* Finishing an infection in the crypt */
/***************************************/

void Crypt::StopInfection() {
	isInfected=false;
	statusLength=0;
}


/*****************************************/
/* Simulating one time step in the crypt */
/*****************************************/

void Crypt::Iteration(int time) {
	// Updating the infection status
	if(isInfected) {
		if(time == 554400) { // 55 weeks = 554,400 minutes
			StopInfection();
		}
		++statusLength;
	}
	else {
		if(time == 50400) {		// 5 weeks = 50,400 minutes
			StartInfection();
		}
		++statusLength;
	}
	
	// Updating the methylation status
	UpdateMethylationProgCells();
	UpdateMethylationStemCell();
		
	// Updating both "ends" of the crypt
	int timeLimit;
	if(isInfected) timeLimit=(int)randomNumberGenerator.randNorm(time-INFECTED_LIFE_SPAN,15);
	else timeLimit=(int)randomNumberGenerator.randNorm(time-NORMAL_LIFE_SPAN,15);
	if(cells[LIMIT_SIZE-sizeLeft].timeIndicator<timeLimit) {
		cells[LIMIT_SIZE-sizeLeft].Erase();
		--sizeLeft;
	}
	if(cells[LIMIT_SIZE+7+sizeRight].timeIndicator<timeLimit) {
		cells[LIMIT_SIZE+7+sizeRight].Erase();
		--sizeRight;
	}

	// Creation of new cells at the bottom
	bool newCellLeft=false, newCellRight=false;
	if(isInfected) {
		if(randomNumberGenerator.randInt(430000)<=infectedSizeParameter) newCellRight=true;
		if(randomNumberGenerator.randInt(430000)<=infectedSizeParameter) newCellLeft=true;
	}
	else {
		if((sizeRight>1.1*NORMAL_SIZE && randomNumberGenerator.randInt(430000)<=4000)||
		   (sizeRight<0.9*NORMAL_SIZE && randomNumberGenerator.randInt(430000)<=25000)||
		   (sizeRight<=1.1*NORMAL_SIZE && sizeRight>=0.9*NORMAL_SIZE &&
		    randomNumberGenerator.randInt(430000)<=10000)) newCellRight=true;
		if((sizeLeft>1.1*NORMAL_SIZE && randomNumberGenerator.randInt(430000)<=4000)||
		   (sizeLeft<0.9*NORMAL_SIZE && randomNumberGenerator.randInt(430000)<=25000)||
		   (sizeLeft<=1.1*NORMAL_SIZE && sizeLeft>=0.9*NORMAL_SIZE &&
		    randomNumberGenerator.randInt(430000)<=10000)) newCellLeft=true;
	}
	if(newCellRight && sizeRight!=LIMIT_SIZE) NewCellRight(time);
	if(newCellLeft && sizeLeft!=LIMIT_SIZE) NewCellLeft(time);	

	// Updating progenitor cells
	// (they can only produce a limited number of diff. cells)
	for(int i=0;i<6;++i) {
		if(cells[progenitorCells[i]].timeIndicator>26) {
			cells[progenitorCells[i]].timeIndicator=0;
			cells[progenitorCells[i]].methylated=cells[stemCell].methylated;
			++(cells[stemCell].timeIndicator);
		}
	}

	// Updating methylation level of the crypt
	int methylatedCells=0;
	for(int i=LIMIT_SIZE-sizeLeft; i<LIMIT_SIZE+7+sizeRight; ++i) {
		if(cells[i].methylated)
		  ++methylatedCells;
	}
	methylationLevel = 100 * (double)methylatedCells / (double)(sizeLeft+7+sizeRight); 

	// Iteration completed
}


/***************************************************/
/* Updating methylation status of progenitor cells */
/***************************************************/

void Crypt::UpdateMethylationProgCells() {
	for(int i=0; i<6; ++i) {
		if(cells[progenitorCells[i]].methylated == false) {
			if(isInfected)
				cells[progenitorCells[i]].methylated=METH_PROG_INF;
			else
				cells[progenitorCells[i]].methylated=METH_PROG_NORM;
		}
	}
}


/************************************************/
/* Updating methylation status of the stem cell */
/************************************************/

void Crypt::UpdateMethylationStemCell() {
	if(cells[stemCell].methylated == false) {
		if(isInfected)
			cells[stemCell].methylated=METH_STEM_INF;
		else
			cells[stemCell].methylated=METH_STEM_NORM;
	}
}


/*******************************************************/
/* Creating a new cell on the left "wall" of the crypt */
/*******************************************************/

void Crypt::NewCellLeft(int time) {
	MoveLeftCells();
	int whichProgCell=randomNumberGenerator.randInt(2);
	whichProgCell=progenitorCells[whichProgCell];
	cells[LIMIT_SIZE-1].timeIndicator=time;
	cells[LIMIT_SIZE-1].cellType=(CellType)2;
	cells[LIMIT_SIZE-1].methylated=cells[whichProgCell].methylated;
	++(cells[whichProgCell].timeIndicator);
	++sizeLeft;
}


/*******************************************/
/* Moving cells upwards on the left "wall" */
/*******************************************/

void Crypt::MoveLeftCells () {
	for(int i=0; i<LIMIT_SIZE-1; ++i) {
		cells[i]=cells[i+1];
	}
}


/********************************************************/
/* Creating a new cell on the right "wall" of the crypt */
/********************************************************/

void Crypt::NewCellRight(int time) {
	MoveRightCells();
	int whichProgCell=3+randomNumberGenerator.randInt(2);
	whichProgCell=progenitorCells[whichProgCell];
	cells[LIMIT_SIZE+7].timeIndicator=time;
	cells[LIMIT_SIZE+7].cellType=(CellType)2;
	cells[LIMIT_SIZE+7].methylated=cells[whichProgCell].methylated;
	++(cells[whichProgCell].timeIndicator);
	++sizeRight;
}


/********************************************/
/* Moving cells upwards on the right "wall" */
/********************************************/

void Crypt::MoveRightCells () {
	for(int i=cellPopulation-1; i>cellPopulation-1-LIMIT_SIZE; --i) {
		cells[i]=cells[i-1];
	}
}


/*****************/
/* Crypt display */
/*****************/

void Crypt::Display() {
	for(int i=0; i<cellPopulation; ++i) {
		cells[i].Display();
	}
}


/***************/
/* End of file */
/***************/
