/* -------------------------------------------------------------------------- */
/* world.cpp                                                                  */
/* -------------------------------------------------------------------------- */

#include "world.hpp"
#include "strain.hpp"
#include "cellule.hpp"
#include "agent.hpp"
#include <time.h>

#include <fstream>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#undef SEEK_SET
#undef SEEK_CUR
#undef SEEK_END

#include "mpi.h"
//#define NBNOEUDS 16

#define STRAIN_NUMBER 500
#define BUFFER_SIZE 3600

// definition of the maximum number of agents in each matrix element
#define NBVIRUSCELL 20
#define NBTHCELL 10
#define NBTCCELL 10
#define NBCPACELL 10


using namespace std;



/***************************************************************
*                                                              *
*    World::World()                                            *
*          constructor                                         *
*                                                              *
***************************************************************/

World::World() {
}


/***************************************************************
*                                                              *
*    World::~World()                                           *
*          destructor                                          *
*                                                              *
***************************************************************/

World::~World() {
}


/***************************************************************
*                                                              *
*    World::init()                                             *
*	   initializes the lymph nodes                             *
*                                                              *
***************************************************************/

int World::init(int inSize,int inNbVirus,int inNbTh,int inNbTc,int inNbCPA) {

   MPI_Comm_rank(MPI_COMM_WORLD,&procNum);
   MPI_Comm_size(MPI_COMM_WORLD,&numberOfNodes);
   
   MPI_Barrier(MPI_COMM_WORLD);

   lymphNode.initStrains();
   lymphNode.procNum=procNum;
   
   // Infection is initiated only in node 2
   if(procNum==2) lymphNode.init(inSize,inNbVirus,inNbTh,inNbTc,inNbCPA);
   else lymphNode.init(inSize,0,inNbTh,inNbTc,inNbCPA);

   nbMovingAgents=0; nbMovingAgents_dep=0; nbMovingAgents_dep2=0;
   return 0 ;
}


/***************************************************************
*                                                              *
*    World::run()                                              *
*	   main function for the world                             *
*          controls lymph node iteration,                      *
*          and communication between them                      *
*                                                              *
***************************************************************/

int World::run() {

   Cellule *ptrCell;
   for(int tempCpt=0; tempCpt<numberOfNodes; ++tempCpt) {
      if(tempCpt==procNum) printf("Run launched... on node %d\n",procNum);
      MPI_Barrier(MPI_COMM_WORLD);
   }

   int i=0;
   int iter_max=50000; // (time step = 47 seconds)
   int buffer[BUFFER_SIZE];

   int t1=0,t2=0,t3=0,t4=0;
   int time_run=0, time_comm=0, time_wait=0, time_temp1, time_temp2, time_temp3, time_temp4, time_temp5, time_deb, time_fin;
   int ptrV,ptrTh,ptrTc,ptrCPA;
   int nbSends=0, nbRecvs=0;
   
   // We use files to save the simulation statistics.
   // Each lymph node uses its own file, called "jobID-procNum.cvs", 
   // where jobID and procNum are replaced by the corresponding values
   FILE *f;
   char filename_[6] = {0};
   char proc_[3];
   sprintf(proc_, "%d", procNum);
   strcat(filename_, jobID);
   strcat(filename_, "-");
   strcat(filename_, proc_);
   strcat(filename_, ".cvs");

   // In each file, we save, for the correspoding lymph node and at several time points:
   //    - the lymph node ID
   //    - the current iteration number
   //    - the number of active virions
   //    - the number of virions locally created since the start of the simulation
   //    - the number of virions locally destroyed since the start of the simulation
   //    - the number of virions sent from the node since the start of the simulation
   //    - the number of virions arrived to the node since the start of the simulation
   //    - the number of active CD4 cells
   //    - the number of CD4 cells locally infected since the start of the simulation
   //    - the number of CD4 cells locally activated since the start of the simulation
   //    - the number of CD4 cells locally created since the start of the simulation
   //    - the number of CD4 cells locally destroyed since the start of the simulation
   //    - the number of CD4 cells sent from the node since the start of the simulation
   //    - the number of CD4 cells arrived to the node since the start of the simulation
   //    - the number of active CD8 cells
   //    - the number of CD8 cells locally activated since the start of the simulation
   //    - the number of CD8 cells locally created since the start of the simulation
   //    - the number of CD8 cells locally destroyed since the start of the simulation
   //    - the number of CD8 cells sent from the node since the start of the simulation
   //    - the number of CD8 cells locally arrived to the node since the start of the simulation
   //    - the number of active APC cells
   //    - the number of APC presentation locally initiated since the start of the simulation
   //    - the number of APC cells locally created since the start of the simulation
   //    - the number of APC cells locally destroyed since the start of the simulation
   //    - the number of APC cells sent from the node since the start of the simulation
   //    - the number of APC cells locally arrived to the node since the start of the simulation
   f = fopen(filename_, "a");
   fprintf(f,"Simulation starting...\nNode\tIter\t");
   fprintf(f,"V\tV-Crea\tV-Kl\tV-Rcv\tV-Snt\t");
   fprintf(f,"Th\tTh-Inf\tTh-Act\tTh-Crea\tTh-Kl\tTh-Rcv\tTh-Snt\t");
   fprintf(f,"Tc\tTc-Act\tTc-Crea\tTc-Kl\tTc-Rcv\tTc-Snt\t");
   fprintf(f,"APC\tAPC-Pr\tAPC-Cr\tAPC-Kl\tAPC-Rv\tAPC-St\n");
   fclose(f);
   MPI_Barrier(MPI_COMM_WORLD);
      
   // Initializing the lymph network
   if(procNum==0) {
      ifstream networkFile;
      networkFile.open("lymphNetwork.dat", ios::in);
      
      // we leave the program if there was an error when opening the file...   
      if(networkFile.bad()) return(1);
      
      // we read the parameters from the file
      int nodeID, rec1, rec2, send1, send2, tempValue, colorValue;
      int path[50];
      networkFile >> tempValue;
      networkFile >> nodeID;
      networkFile >> rec1;
      networkFile >> rec2;
      networkFile >> send1;
      networkFile >> send2;
      int pathIndex=0;
      do {
         networkFile >> tempValue;
         path[pathIndex]=tempValue;
         ++pathIndex;
      } while(tempValue!=-1);
      networkFile >> colorValue;
      recvFrom[0]=rec1;recvFrom[1]=rec2;
      sendTo[0]=send1;sendTo[1]=send2;
      //recvFromBackup[0]=recvFrom[0];recvFromBackup[1]=recvFrom[1];
      //sendToBackup[0]=sendTo[0];sendToBackup[1]=sendTo[1];
      pathIndex=0;
      do {
         tempValue=path[pathIndex];
         route[pathIndex]=tempValue;
         ++pathIndex;
      } while(tempValue!=-1);
      color=colorValue;
      // Now node0 is ready
      printf("Node %d has accepted network generation...\n", procNum);   

      // Now it reads, and send to the others

      for(int destNode=1; destNode<numberOfNodes; ++destNode) {
         networkFile >> nodeID;
         networkFile >> rec1;
         networkFile >> rec2;
         networkFile >> send1;
         networkFile >> send2;
         int pathIndex=0;
         do {
            networkFile >> tempValue;
            path[pathIndex]=tempValue;
            ++pathIndex;
         } while(tempValue!=-1);
         networkFile >> colorValue;
         int tempBuffer[6];
         tempBuffer[0]=nodeID;tempBuffer[1]=rec1;tempBuffer[2]=rec2;
         tempBuffer[3]=send1;tempBuffer[4]=send2;tempBuffer[5]=colorValue;
         MPI_Send(&tempBuffer,6,MPI_INT,destNode,1,MPI_COMM_WORLD);
         MPI_Send(&path,50,MPI_INT,destNode,1,MPI_COMM_WORLD);
      }
   }
   else { // Other nodes only receives info
      int tempBuffer[6];
      int path[50];
      MPI_Recv(&tempBuffer,6,MPI_INT,0,1,MPI_COMM_WORLD,MPI_STATUS_IGNORE);
      MPI_Recv(&path,50,MPI_INT,0,1,MPI_COMM_WORLD,MPI_STATUS_IGNORE);
      
      // We just make sure the network generation file is not corrupted
      if(procNum!=tempBuffer[0]) printf("there is problem with node %d (%d)\n",procNum,tempBuffer[0]);

      recvFrom[0]=tempBuffer[1];recvFrom[1]=tempBuffer[2];
      sendTo[0]=tempBuffer[3];sendTo[1]=tempBuffer[4];
      int pathIndex=0;
      int tempValue;
      do {
         tempValue=path[pathIndex];
         route[pathIndex]=tempValue;
         ++pathIndex;
      } while(tempValue!=-1);
      color=tempBuffer[5];
      printf("Node %d has accepted network generation...\n", procNum);   
   }

   MPI_Barrier(MPI_COMM_WORLD);
   if(procNum==0) printf("Network generation is completed...\n");

   time_deb=time(0);
   
   MPI_Barrier(MPI_COMM_WORLD);
   if(procNum==0) printf("Simulation is starting...\n");

   // Main loop
   while(i<iter_max) {

      // Iterations
      
      time_temp1=time(0);
      lymphNode.iteration(i,this,procNum); ++i;

      // Here, we want to save extra details on the first time steps
      // This can be commented out
      if(i<300) {
         f = fopen(filename_, "a");
         fprintf(f,"%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t\n",
                    procNum, i, lymphNode.nbV_, lymphNode.nbVCrea, lymphNode.nbVKill, lymphNode.nbVRec, lymphNode.nbVSent,
                    lymphNode.nbTh_, lymphNode.nbThInf, lymphNode.nbThAct, lymphNode.nbThCrea, lymphNode.nbThKill, 
                    lymphNode.nbThRec, lymphNode.nbThSent, lymphNode.nbTc_, lymphNode.nbTcAct, lymphNode.nbTcCrea, 
                    lymphNode.nbTcKill, lymphNode.nbTcRec, lymphNode.nbTcSent, lymphNode.nbCPA_, lymphNode.nbCPAPres, 
                    lymphNode.nbCPACrea, lymphNode.nbCPAKill, lymphNode.nbCPARec, lymphNode.nbCPASent);

         lymphNode.iteration(i,this,procNum); ++i;

         fprintf(f,"%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t\n",
                    procNum, i, lymphNode.nbV_, lymphNode.nbVCrea, lymphNode.nbVKill, lymphNode.nbVRec, lymphNode.nbVSent,
                    lymphNode.nbTh_, lymphNode.nbThInf, lymphNode.nbThAct, lymphNode.nbThCrea, lymphNode.nbThKill, 
                    lymphNode.nbThRec, lymphNode.nbThSent, lymphNode.nbTc_, lymphNode.nbTcAct, lymphNode.nbTcCrea, 
                    lymphNode.nbTcKill, lymphNode.nbTcRec, lymphNode.nbTcSent, lymphNode.nbCPA_, lymphNode.nbCPAPres, 
                    lymphNode.nbCPACrea, lymphNode.nbCPAKill, lymphNode.nbCPARec, lymphNode.nbCPASent);

         lymphNode.iteration(i,this,procNum); ++i;

         fprintf(f,"%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t\n",
                    procNum, i, lymphNode.nbV_, lymphNode.nbVCrea, lymphNode.nbVKill, lymphNode.nbVRec, lymphNode.nbVSent,
                    lymphNode.nbTh_, lymphNode.nbThInf, lymphNode.nbThAct, lymphNode.nbThCrea, lymphNode.nbThKill, 
                    lymphNode.nbThRec, lymphNode.nbThSent, lymphNode.nbTc_, lymphNode.nbTcAct, lymphNode.nbTcCrea, 
                    lymphNode.nbTcKill, lymphNode.nbTcRec, lymphNode.nbTcSent, lymphNode.nbCPA_, lymphNode.nbCPAPres, 
                    lymphNode.nbCPACrea, lymphNode.nbCPAKill, lymphNode.nbCPARec, lymphNode.nbCPASent);
         fclose(f);
      }
      else {
         lymphNode.iteration(i,this,procNum); ++i;
         lymphNode.iteration(i,this,procNum); ++i;
      }

      time_temp2=time(0);
      time_run+=(time_temp2-time_temp1);
      
      // Exchanging agent lists
      
      int size_,size2_;
      
      // Color of the nodes is used to limit communication
      // (see Thesis for details)
      
      if(color==(i%2)) {

         size_=9*nbMovingAgents_dep;
         time_temp3=time(0);
         MPI_Send(&size_,1,MPI_INT,sendTo[0],1,MPI_COMM_WORLD);
         time_temp4=time(0);
         time_wait+=(time_temp4-time_temp3);
      
         if(size_) {
            MPI_Send(list_dep,size_,MPI_INT,sendTo[0],2,MPI_COMM_WORLD);
            nbMovingAgents_dep=0;
            ++nbSends;
         }
         if(sendTo[1]>=0) {
            size_=9*nbMovingAgents_dep2;
            time_temp3=time(0);
            MPI_Send(&size_,1,MPI_INT,sendTo[1],1,MPI_COMM_WORLD);
            time_temp4=time(0);
            time_wait+=(time_temp4-time_temp3);
            if(size_) {
               MPI_Send(list_dep2,size_,MPI_INT,sendTo[1],2,MPI_COMM_WORLD);
               nbMovingAgents_dep2=0;
               ++nbSends;
            }
         }
      }

      else {
         time_temp3=time(0);
         nbMovingAgents=0;
         MPI_Recv(&size2_,1,MPI_INT,recvFrom[0],1,MPI_COMM_WORLD,MPI_STATUS_IGNORE);
         time_temp4=time(0);
         time_wait+=(time_temp4-time_temp3);
         if(size2_) {
            ++nbRecvs;
            MPI_Recv(buffer,size2_,MPI_INT,recvFrom[0],2,MPI_COMM_WORLD,MPI_STATUS_IGNORE);
            for(int c2=0;c2<size2_;c2+=9) {
               addMovingAgent(buffer[c2],buffer[c2+1],buffer[c2+2],buffer[c2+3],
                              buffer[c2+4],buffer[c2+5],buffer[c2+6],buffer[c2+7],buffer[c2+8]);
            }
         }
         if(recvFrom[1]>=0) {
            time_temp3=time(0);
            MPI_Recv(&size2_,1,MPI_INT,recvFrom[1],1,MPI_COMM_WORLD,MPI_STATUS_IGNORE);
            time_temp4=time(0);
            time_wait+=(time_temp4-time_temp3);
            if(size2_) {
               ++nbRecvs;
               MPI_Recv(buffer,size2_,MPI_INT,recvFrom[1],2,MPI_COMM_WORLD,MPI_STATUS_IGNORE);
               for(int c2=0;c2<size2_;c2+=9) {
                  addMovingAgent(buffer[c2],buffer[c2+1],buffer[c2+2],buffer[c2+3],
                                 buffer[c2+4],buffer[c2+5],buffer[c2+6],buffer[c2+7],buffer[c2+8]);
               }
            }
         }
         time_temp3=time(0);
         time_temp4=time(0);
         time_wait+=(time_temp4-time_temp3);

         time_temp5=time(0);
         time_comm+=(time_temp5-time_temp2);

         // Each node now has the list of agents arriving in the local matrix
         int cell_i,cell_j;
         for(int c2=0;c2<9*nbMovingAgents;c2+=9) {
            cell_i=0;cell_j=0;
            switch(list[c2+1]) {
               case 0:
                  do {
                     ptrCell = lymphNode.access(cell_i,cell_j);
                     ++cell_i;
                     if(cell_i==lymphNode.size_) {cell_i=0; ++cell_j;}
                  }
                  while(ptrCell->getNbcV()==NBVIRUSCELL);
                  ptrV = lymphNode.V_free.back();
                  lymphNode.V_free.pop_back();
                  ptrCell->tabV[ptrCell->getNbcV()]=ptrV;
                  if(list[c2]==procNum) lymphNode.V_tab[ptrV].setDest(-1); // agent has arrived
                  else lymphNode.V_tab[ptrV].setDest(list[c2]);
                  lymphNode.V_tab[ptrV].setDate(list[c2+2]);
                  lymphNode.V_tab[ptrV].setMaxAge(list[c2+3]);
                  lymphNode.V_tab[ptrV].setStrain(list[c2+4]);
                  ptrCell->nbcVInc(); lymphNode.nbV_++;  lymphNode.nbVRec++;
                  break;
               case 1:
                  do {
                     ptrCell = lymphNode.access(cell_i,cell_j);
                     ++cell_i;
                     if(cell_i==lymphNode.size_) {cell_i=0; ++cell_j;}
                  }
                  while(ptrCell->getNbcTh()==NBTHCELL);
                  ptrTh = lymphNode.Th_free.back();
                  lymphNode.Th_free.pop_back();
                  ptrCell->tabTh[ptrCell->getNbcTh()]=ptrTh;
                  if(list[c2]==procNum) lymphNode.Th_tab[ptrTh].setDest(-1);
                  else lymphNode.Th_tab[ptrTh].setDest(list[c2]);
                  lymphNode.Th_tab[ptrTh].setDate(list[c2+2]);
                  lymphNode.Th_tab[ptrTh].setMaxAge(list[c2+3]);
                  lymphNode.Th_tab[ptrTh].setClonotype(list[c2+4]);
                  lymphNode.Th_tab[ptrTh].setActive(list[c2+5]);
                  lymphNode.Th_tab[ptrTh].setInfected(list[c2+6]);
                  lymphNode.Th_tab[ptrTh].setMemory(list[c2+7]);
                  lymphNode.Th_tab[ptrTh].setExpansion(list[c2+8]);
                  ptrCell->nbcThInc(); lymphNode.nbTh_++;  lymphNode.nbThRec++;
                  break;
               case 2:
                  do {
                     ptrCell = lymphNode.access(cell_i,cell_j);
                     ++cell_i;
                     if(cell_i==lymphNode.size_) {cell_i=0; ++cell_j;}
                  }
                  while(ptrCell->getNbcTc()==NBTCCELL);
                  ptrTc = lymphNode.Tc_free.back();
                  lymphNode.Tc_free.pop_back();
                  ptrCell->tabTc[ptrCell->getNbcTc()]=ptrTc;
                  if(list[c2]==procNum) lymphNode.Tc_tab[ptrTc].setDest(-1);
                  else lymphNode.Tc_tab[ptrTc].setDest(list[c2]);
                  lymphNode.Tc_tab[ptrTc].setDate(list[c2+2]);
                  lymphNode.Tc_tab[ptrTc].setMaxAge(list[c2+3]);
                  lymphNode.Tc_tab[ptrTc].setClonotype(list[c2+4]);
                  lymphNode.Tc_tab[ptrTc].setActive(list[c2+5]);
                  lymphNode.Tc_tab[ptrTc].setExpansion(list[c2+6]);
                  lymphNode.Tc_tab[ptrTc].setMemory(list[c2+7]);
                  ptrCell->nbcTcInc(); lymphNode.nbTc_++;  lymphNode.nbTcRec++;
                  break;
               case 3:
                  do {
                     ptrCell = lymphNode.access(cell_i,cell_j);
                     ++cell_i;
                     if(cell_i==lymphNode.size_) {cell_i=0; ++cell_j;}
                  }
                  while(ptrCell->getNbcCPA()==NBCPACELL);
                  ptrCPA = lymphNode.CPA_free.back();
                  lymphNode.CPA_free.pop_back();
                  ptrCell->tabCPA[ptrCell->getNbcCPA()]=ptrCPA;
                  if(list[c2]==procNum) lymphNode.CPA_tab[ptrCPA].setDest(-1);
                  else lymphNode.CPA_tab[ptrCPA].setDest(list[c2]);
                  lymphNode.CPA_tab[ptrCPA].setDate(list[c2+2]);
                  lymphNode.CPA_tab[ptrCPA].setMaxAge(list[c2+3]);
                  lymphNode.CPA_tab[ptrCPA].setPresenting1(list[c2+4]);
                  lymphNode.CPA_tab[ptrCPA].setPresenting2(list[c2+5]);
                  lymphNode.CPA_tab[ptrCPA].setPresenting3(list[c2+6]);
                  lymphNode.CPA_tab[ptrCPA].setPresenting4(list[c2+7]);
                  lymphNode.CPA_tab[ptrCPA].setPresenting5(list[c2+8]);
                  ptrCell->nbcCPAInc(); lymphNode.nbCPA_++;  lymphNode.nbCPARec++;
                  break;
            }
         }
         nbMovingAgents=0;
      }

      if((i%100)==0) { // We show the simulation progress
                        //  (every 300 iterations, since iterations are grouped by 3 above)
         if(procNum==0) printf("Iteration %d completed\n",i);
      }


      if((i%1800)==0) { // We save the simulation status every 1800 iterations (almost 24h of "real time")
         f = fopen(filename_, "a");
         fprintf(f,"%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t\n",
                 procNum,i,
                 lymphNode.nbV_,lymphNode.nbVCrea, lymphNode.nbVKill, lymphNode.nbVRec, lymphNode.nbVSent,
                 lymphNode.nbTh_,lymphNode.nbThInf,lymphNode.nbThAct,lymphNode.nbThCrea, lymphNode.nbThKill, lymphNode.nbThRec, lymphNode.nbThSent,
                 lymphNode.nbTc_,lymphNode.nbTcAct,lymphNode.nbTcCrea, lymphNode.nbTcKill, lymphNode.nbTcRec, lymphNode.nbTcSent,
                 lymphNode.nbCPA_,lymphNode.nbCPAPres,lymphNode.nbCPACrea, lymphNode.nbCPAKill, lymphNode.nbCPARec, lymphNode.nbCPASent);
         fclose(f);
      }
   }  // End of the main loop
   time_fin=time(0);

   printf("\nit's over\n");

   t1=0;t2=0;t3=0;t4=0;
   for(int cpt=0;cpt<lymphNode.size_*lymphNode.size_;++cpt) {
      ptrCell=&(lymphNode.myMatrix_[cpt]);
      t1+=ptrCell->getNbcV();
      t2+=ptrCell->getNbcTh();
      t3+=ptrCell->getNbcTc();
      t4+=ptrCell->getNbcCPA();
   }

   printf("\n*** NODE %d ***\n",procNum);
   return i;
}



/***************************************************************
*                                                              *
*    World::addMovingAgent()                                   *
*          adds an agent to the list                           *
*                                                              *
***************************************************************/

int World::addMovingAgent(int dest, int type, int date, int age, int p1, int p2, int p3, int p4, int p5) {
   int returnValue=1;
   if(nbMovingAgents+9<BUFFER_SIZE) {
      list[9*nbMovingAgents] = dest;
      list[9*nbMovingAgents+1] = type;
      list[9*nbMovingAgents+2] = date;
      list[9*nbMovingAgents+3] = age;
      list[9*nbMovingAgents+4] = p1;
      list[9*nbMovingAgents+5] = p2;
      list[9*nbMovingAgents+6] = p3;
      list[9*nbMovingAgents+7] = p4;
      list[9*nbMovingAgents+8] = p5;
      ++nbMovingAgents;
      returnValue=0;
   }
   return returnValue;
}



/***************************************************************
*                                                              *
*    World::addMovingAgent_list1()                             *
*          adds an agent to the list                           *
*                                                              *
***************************************************************/

int World::addMovingAgent_list1(int dest, int type, int date, int age, int p1, int p2, int p3, int p4, int p5) {
   int returnValue=1;
   if(nbMovingAgents_dep+9<BUFFER_SIZE) {
      list_dep[9*nbMovingAgents_dep] = dest;
      list_dep[9*nbMovingAgents_dep+1] = type;
      list_dep[9*nbMovingAgents_dep+2] = date;
      list_dep[9*nbMovingAgents_dep+3] = age;
      list_dep[9*nbMovingAgents_dep+4] = p1;
      list_dep[9*nbMovingAgents_dep+5] = p2;
      list_dep[9*nbMovingAgents_dep+6] = p3;
      list_dep[9*nbMovingAgents_dep+7] = p4;
      list_dep[9*nbMovingAgents_dep+8] = p5;
      ++nbMovingAgents_dep;
      returnValue=0;
   }
   return returnValue;
}



/***************************************************************
*                                                              *
*    World::addMovingAgent_list2()                             *
*          adds an agent to the list                           *
*                                                              *
***************************************************************/

int World::addMovingAgent_list2(int dest, int type, int date, int age, int p1, int p2, int p3, int p4, int p5) {
   int returnValue=1;
   if(nbMovingAgents_dep2+9<BUFFER_SIZE) {
      list_dep2[9*nbMovingAgents_dep2] = dest;
      list_dep2[9*nbMovingAgents_dep2+1] = type;
      list_dep2[9*nbMovingAgents_dep2+2] = date;
      list_dep2[9*nbMovingAgents_dep2+3] = age;
      list_dep2[9*nbMovingAgents_dep2+4] = p1;
      list_dep2[9*nbMovingAgents_dep2+5] = p2;
      list_dep2[9*nbMovingAgents_dep2+6] = p3;
      list_dep2[9*nbMovingAgents_dep2+7] = p4;
      list_dep2[9*nbMovingAgents_dep2+8] = p5;
      ++nbMovingAgents_dep2;
      returnValue=0;
   }
   return returnValue;
}

/***************************************************************
*                                                              *
*    World::appartient()                                       *
*          checks whether the destination ("val") is among     *
*          the first six nodes of a side-chain ("route")       *
*                                                              *
***************************************************************/

bool World::appartient(int* route, int val) {
   if(val==route[0]) return true;
   if(val==route[1]) return true;
   if(val==route[2]) return true;
   if(val==route[3]) return true;
   if(val==route[4]) return true;
   if(val==route[5]) return true;
   return false;
}



/* ------------------------------ End Of File ------------------------------- */
