#ifndef __INCLUDES_H__
#define __INCLUDES_H__

#include <stdbool.h>

typedef unsigned       uint16;
typedef unsigned long  uint32;

typedef enum {PROMOTE, DEMOTE, LS} Method;

struct _Solution
{
  uint16 size; /* number of nodes of the graph */
  bool *vect;  /* decision vector on the nodes */
  int  value;  /* value of the solution        */
  uint16 nb_active_c; // nb active conditions
  uint16 nb_active_g; // nb active genes
};
typedef struct _Solution Solution;


typedef struct _DicoRoot
{
  struct _DicoNode * fils0;
  struct _DicoNode * fils1;
} DicoRoot;


typedef struct _DicoNode
{
  uint16             ind;
  bool               val;
  struct _DicoNode * fils0;
  struct _DicoNode * fils1;
} DicoNode;


typedef struct _DicoLeaf
{
  uint16             ind;
  bool               val;
  Solution         * sol;
} DicoLeaf;


typedef struct _DicoSol
{
  Solution * sol;
  Solution * final;
} DicoSol;


typedef struct _DicoTable
{
  DicoRoot  *root;
  uint16    size;
  uint16    nb_sol;
  Solution  *sols;
} DicoTable;

typedef struct _Path
{
  uint16 id;
} Path;


struct _Chromosome
{
  uint16 size; /* number of conditions: set B          */
  bool *conds;  /* decision vector on the condtions    */
  int nb_active_g; /* number of genes in solution      */
  int nb_active_c; /* number of conditions in solution */
  int  value;  /* value of the chromosome              */
};
typedef struct _Chromosome Chromosome;


typedef struct
{
  uint16   size; // nb conditions = size of the Pareto front
  Solution * vec; // nb conditions - indexed Pareto front
  uint16   * nb_upd; // nb improvements for each point
  uint16   * nb_tst; // nb tries for each point
} Front;


struct _Problem
{
  char   name[64]; /* name of the instance                  */
  uint16 nb_nodes; /* number of nodes: sets A and B         */
  uint16 nb_genes; /* number of genes: set A                */
  uint16 nb_cond;  /* nomber of conditions: set B           */
  int    **C;      /* cost matrix: complete bipartite graph */

  bool *used;   /* decision vector on the nodes           */
  bool *used_g; /* decision vector on the genes           */
  bool *used_c; /* decision vector on the conditions      */
  int  *cost;   /* add/drop cost vector on the nodes      */
  int  *cost_g; /* add/drop cost vector on the genes      */
  int  *cost_c; /* add/drop cost vector on the conditions */
  int  value;   /* value of the solution */
  uint16 nb_active_c; // nb active conditions
  uint16 nb_active_g; // nb active genes
};
typedef struct _Problem Problem;


/* ====================================================================== */
/* helpful macros                                                         */
/* ====================================================================== */

#define ALLOCATION(nb,type) ((type*)malloc((nb)*sizeof(type)))
#define FREE(x) if ((x) != NULL) free(x)

#define ENTER printf("Entering %s\n", __PRETTY_FUNCTION__); fflush(stdout);
#define LEAVE printf("Leaving %s\n", __PRETTY_FUNCTION__); fflush(stdout);


/* ====================================================================== */
/* global setting                                                         */
/* ====================================================================== */

#undef NDEBUG


/* ====================================================================== */
/* random number generator: Mersenne Twister                              */
/* ====================================================================== */

int    Uniform (int min, int max);
double randExc (void);
double randDblExc (void);
double randNorm (const double mean, const double variance);
void   seedMT (uint32 seed);
uint32 reloadMT (void);
uint32 randomMT (void);


/* ====================================================================== */
/* utilities                                                              */
/* ====================================================================== */

void error (char const file[], uint16 line, char const func[], char const msg[], ...);
double startChrono ();
double stopChrono (double start);
#define ERROR(...) error(__FILE__, __LINE__, __func__, __VA_ARGS__)


/* ====================================================================== */
/* problem manipulation functions                                         */
/* ====================================================================== */

void allocProblem (Problem * pb, uint16 nb_g, uint16 nb_c);
void freeProblem (Problem * pb);
void clearGraphSolution (Problem * pb);
void randomProblem1 (Problem * pb);
void randomProblem2 (Problem * pb);
void readProblem (Problem * pb, const char filename[]);
void writeProblem (Problem * pb, const char filename[]);
void displayProblem (Problem * pb);
void displayGraphSolution (Problem * pb, bool reduced);
void checkGraphSolution (Problem * pb);
void reduceProblem (Problem * pb);


/* ====================================================================== */
/* solution-based functions                                               */
/* ====================================================================== */

void allocSolution (Solution * s, uint16 size);
void freeSolution (Solution * s);
void clearSolution (Solution * s);
void copySolution (Solution * s1, Solution * s2);
void randomSolution (Solution * s, Problem * pb);
void displaySolution (Solution * s, bool detailled);
void saveSolution (Solution * s, const char filename[]);
void extractSolution (Problem * pb, Solution * s);
void insertSolution (Problem * pb, Solution * s);


/* ====================================================================== */
/* dictionary-based functions                                             */
/* ====================================================================== */

void CreateDico (DicoTable * d, uint16 size_sol, uint16 max_sols);
void DestroyDico (DicoTable * d);
void CheckDico (DicoTable * d);
void DisplayDicoSol (DicoSol * s);
void DisplayDico (DicoTable * d);
Path * InsertDico (DicoTable * d, Solution * s);
bool IsInDico (DicoTable * d, Solution * s);
void ActivatePath (Path * p, uint16 l);


/* ====================================================================== */
/* exact methods                                                          */
/* ====================================================================== */

void enumerate  (Problem * pb, Solution * s);
void enumerate2 (Problem * pb, bool save);
void writeIP    (Problem * pb, const char name[]);
void writeQP    (Problem * pb, const char name[]);


/* ====================================================================== */
/* basic local search                                                     */
/* ====================================================================== */

void changeState (Problem * pb, uint16 node);
void construction1 (Problem * pb, float alpha);
void construction2 (Problem * pb, float alpha);
void LS1 (Problem * pb, Solution * s);
void LS2 (Problem * pb, Solution * s);
void VND (Problem * pb);
void GRASP (Problem * pb, uint16 nb_it, Solution * s);
void Tabou (Problem * pb, uint16 nb_it, Solution * s);

void bestNoCond     (Problem * pb, Solution * s);
void best1Cond      (Problem * pb, Solution * s);
void best2Cond      (Problem * pb, Solution * s);
void best3Cond      (Problem * pb, Solution * s);
void bestAllCond    (Problem * pb, Solution * s);
void bestMinus1Cond (Problem * pb, Solution * s);
void bestMinus2Cond (Problem * pb, Solution * s);
void bestMinus3Cond (Problem * pb, Solution * s);

bool promote        (Problem * pb, Solution * s);
bool demote         (Problem * pb, Solution * s);


/* ====================================================================== */
/* other local searches                                                   */
/* ====================================================================== */

void LS3 (Problem * pb, Solution * s);
void LS4 (Problem * pb, Solution * s);
void randomSolutionCard (Solution * s, Problem * pb, uint16 card);
void LS5 (Problem * pb, Solution * s);
void LS6 (Problem * pb, Solution * s);



/* ====================================================================== */
/* genetic algorithm                                                      */
/* ====================================================================== */

void allocChromosome (Chromosome * ch, uint16 size);
void freeChromosome (Chromosome * ch);
void randomChromosome (Chromosome * ch);
void evalChromosome (Problem * pb, Chromosome * ch);
void eval2Chromosome (Problem * pb, Chromosome * ch);
void displayChromosome (Chromosome * ch);
void allocPopulation (Chromosome * ch, uint16 populationSize, uint16 chromosomeSize);
void freePopulation (Chromosome * ch, uint16 populationSize);
void genetic (Problem * pb, uint16 populationSize, Solution *sol, bool save);
void simpleMutation (Chromosome * ch);
void crossOver1 (Chromosome * ch1, Chromosome * ch2);
void crossOver2 (Chromosome * ch1, Chromosome * ch2);
void localSearch (Problem * pb, Chromosome * ch);
void chromosomeToSolution (Problem * pb, Chromosome * ch, Solution * sol);
void solutionToChromosome (Problem * pb, Chromosome * ch, Solution * sol);
void diversification (Chromosome * ch, Chromosome * population, uint16 populationSize);
void sortPopulation (Chromosome * population, uint16 populationSize);
void swapChromosomes (Chromosome * population, uint16 ch1, uint16 ch2);
uint16 removeRedundancy (Chromosome * population, uint16 populationSize);
bool identicalChromosomes (Chromosome * ch1, Chromosome * ch2);
void removeChromosome (Chromosome * population, uint16 populationSize, uint16 position);
void copyChromosome (Chromosome * ch_source, Chromosome * ch_dest);
void displayChromosome (Chromosome * ch);
void displayChromosomeShort (Chromosome * ch);
void displayChromosomeLong (Chromosome * ch, Problem * pb);
void initPopulation (Problem * pb, Chromosome * population, uint16 populationSize);
void initOrthoPopulation (Problem * pb, Chromosome * population, uint16 populationSize);
void initOrthoPopulation2 (Problem * pb, Chromosome * population, uint16 populationSize);


/* ====================================================================== */
/* main function                                                          */
/* ====================================================================== */

int main (int argc, char * argv[]);

#endif
