#include "defines.h"

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>


/* ====================================================================== */
/* check all the solutions: assume nb_cond << nb_gene                     */
/* ====================================================================== */

void enumerate (Problem * pb, Solution * s)
{
  int value, val[pb->nb_genes];

  // initialization
  clearSolution (s);
  for (uint16 j = 0; j < pb->nb_cond; j++)
    pb->used_c[j] = false;

  // enumeration loop
  uint16 full_cond = (1 << pb->nb_cond) - 1;
  for (uint16 cond = 1; cond <= full_cond; cond++)
  {
    // set up the value vector
    for (uint16 i = 0; i < pb->nb_genes; i++)
      val[i] = 0;

    // set up the cond vector
    uint16 v = cond;
    uint16 elt = 0;
    do
    {
      // set up the use of the condition
      pb->used_c[elt] = v & 1;

      // update the add cost of each gene
      if (pb->used_c[elt] == true)
      {
        for (uint16 i = 0; i < pb->nb_genes; i++)
          val[i] += pb->C[i][elt];
      }

      v >>= 1;
      elt++;
    } while (v > 0);

    // compute the best set of genes
    value = 0;
    for (uint16 i = 0; i < pb->nb_genes; i++)
    {
      bool tst = (val[i] < 0);
      pb->used[i] = tst;
      if (tst)
        value += val[i];
    }
    pb->value = value;
    if (pb->value < s->value)
      extractSolution (pb, s);
  }

  // check the best solution
  insertSolution (pb, s);
  checkGraphSolution (pb);
}


/* ====================================================================== */
/* check all the solutions: assume nb_cond << nb_gene                     */
/* version 2: store the best solution for each nb of conditions           */
/* ====================================================================== */

void enumerate2 (Problem * pb, bool save)
{
  int value, val[pb->nb_genes];
  Solution sol[pb->nb_cond + 1];

  // initialization
  for (uint16 j = 0; j < pb->nb_cond; j++)
    pb->used_c[j] = false;
  for (uint16 j = 0; j <= pb->nb_cond; j++)
  {
    allocSolution (sol + j, pb->nb_nodes);
    clearSolution (sol + j);
    sol[j].value = INT_MAX;
  }
  sol[0].value = 0;

  // enumeration loop
  uint16 full_cond = (1 << pb->nb_cond) - 1;
  for (uint16 cond = 1; cond <= full_cond; cond++)
  {
    // set up the value vector
    for (uint16 i = 0; i < pb->nb_genes; i++)
      val[i] = 0;

    // set up the cond vector
    uint16 v = cond;
    uint16 elt = 0;
    uint16 card = 0;
    do
    {
      // set up the use of the condition
      pb->used_c[elt] = v & 1;

      // update the add cost of each gene
      if (pb->used_c[elt] == true)
      {
        for (uint16 i = 0; i < pb->nb_genes; i++)
          val[i] += pb->C[i][elt];
        card++;
      }

      v >>= 1;
      elt++;
    } while (v > 0);

    // compute the best set of genes
    value = 0;
    for (uint16 i = 0; i < pb->nb_genes; i++)
    {
      bool tst = (val[i] < 0);
      pb->used[i] = tst;
      if (tst)
        value += val[i];
    }
    pb->value = value;
    if (value < sol[card].value)
      extractSolution (pb, sol + card);
  }

  // check the solutions
  for (uint16 j = 0; j < pb->nb_cond + 1; j++)
  {
    uint16 card = 0;
//    insertSolution (pb, sol + j);
//    checkGraphSolution (pb);
    for (uint16 i = 0; i < sol[j].size; i++)
      card += sol[j].vect[i];
    printf("%2d cond: %5d genes, val = %8d\n",
      j, card - j, sol[j].value);
  }

  // dump to a gnuplot file
  if (save)
  {
    int ymax, ymin;
    char cmd[64];
    const char datafile[] = "enum.dat";
    const char outputfile[] = "enum.eps";
    const char cmdfile[]  = "display.p";
    FILE * handle;

    // open the data file
    if ((handle = fopen(datafile, "wt")) == NULL)
      ERROR("cannot open file %s\n", datafile);

    // write the data
    fprintf(handle, "# nb_conditions / value\n\n");
    ymax = INT_MIN; ymin = INT_MAX;
    for (uint16 j = 1; j < pb->nb_cond + 1; j++)
    {
      fprintf(handle, "%d %d\n", j, sol[j].value);
      if (sol[j].value > ymax)
        ymax = sol[j].value;
      if (ymin > sol[j].value)
        ymin = sol[j].value;
    }
    fprintf(handle, "\n");

    // close the file
    fclose(handle);

    // write down the command file
    if ((handle = fopen(cmdfile, "wt")) == NULL)
      ERROR("cannot open file %s\n", cmdfile);

    // write the data
    fprintf(handle, "set encoding iso_8859_1\n");
    fprintf(handle, "unset log\nunset label\n");
    fprintf(handle, "set title \"impact of nb active conditions (%s)\"\n",
      pb->name);
    fprintf(handle, "set xlabel \"nb conditions\"\n");
    fprintf(handle, "set ylabel \"optimal value\"\n");
    fprintf(handle, "set xrange [%g:%g]\n", 0.0, pb->nb_cond + 0.5);
    for (uint16 j = 1; j < pb->nb_cond + 1; j++)
    {
      uint16 card = 0;
      for (uint16 i = 0; i < sol[j].size; i++)
        card += sol[j].vect[i];
      fprintf(handle, "set label \'%d\' at %g,%g\n",
        card - j, 0.1 + (float)(j), (float)((ymin - ymax)/40 + sol[j].value));
    }
    fprintf(handle, "plot \'%s\' using 1:2 notitle with linespoints\n",
      datafile);
    fprintf(handle, "set term postscript eps\n");
    fprintf(handle, "set output \'%s\'\n", outputfile);
    fprintf(handle, "replot\n");
    fprintf(handle, "\n");

    // close the file
    fclose(handle);

    sprintf(cmd, "gnuplot %s", cmdfile);
    system(cmd);
  }

  // free the memoyy
  for (uint16 j = 0; j <= pb->nb_cond; j++)
  {
    freeSolution (sol + j);
  }
}


/* ====================================================================== */
/* write the IP-model of the problem                                      */
/* ====================================================================== */

void writeIP (Problem * pb, const char name[])
{
  FILE * handle;
  uint16 cpt;

  // open the output file
  handle = fopen(name, "wt");

  // write the preamble

  // write the objective function
  fprintf(handle, "Minimize\ncost:");
  cpt = 0;
  for (uint16 i = 0; i < pb->nb_genes; i++)
  {
    for (uint16 j = 0; j < pb->nb_cond; j++)
    {
      cpt += fprintf(handle, " %+d X%d_%d", pb->C[i][j], i, j);
      if (cpt > 80)
      {
        fprintf(handle, "\n");
        cpt = 0;
      }
    }
  }
  fprintf(handle, "\n");

  // write the constraints
  fprintf(handle, "\nSubject to\n");
  for (uint16 i = 0; i < pb->nb_genes; i++)
  {
    for (uint16 j = 0; j < pb->nb_cond; j++)
    {
      fprintf(handle, "\n");
      fprintf(handle, "arc_%d_%d_%d: X%d_%d - Y%d <= 0\n", i, j, 0, i, j, i);
      fprintf(handle, "arc_%d_%d_%d: X%d_%d - Z%d <= 0\n", i, j, 1, i, j, j);
      fprintf(handle, "arc_%d_%d_%d: X%d_%d - Y%d - Z%d >= -1\n", i, j, 2, i, j, i, j);
    }
  }
  fprintf(handle, "\n");

  // enforcing constraints on the cardinality
/*
  fprintf(handle, "sum_X:");
  for (uint16 i = 0; i < pb->nb_genes; i++)
  {
    for (uint16 j = 0; j < pb->nb_cond; j++)
    {
      fprintf(handle, " + X%d_%d", i, j);
    }
  }
  fprintf(handle, " - X = 0\n");
  fprintf(handle, "sum_Y:");
  for (uint16 i = 0; i < pb->nb_genes; i++)
  {
    fprintf(handle, " + Y%d", i);
  }
  fprintf(handle, " - Y = 0\n");
  fprintf(handle, "sum_Z:");
  for (uint16 j = 0; j < pb->nb_cond; j++)
  {
    fprintf(handle, " + Z%d", j);
  }
  fprintf(handle, " - Z = 0\n");
*/
/*
  fprintf(handle, "sum_cstr1: X - %d Y <= 0\n", pb->nb_cond);
  fprintf(handle, "sum_cstr2: X - %d Z <= 0\n", pb->nb_genes);
  fprintf(handle, "sum_cstr3: X - %d Y - %d Z >= -%d\n", pb->nb_cond, pb->nb_genes, pb->nb_genes * pb->nb_cond);
  fprintf(handle, "\n");
*/
  // fprintf(handle, "sum_cstr: X - [ Y * Z ] <= 0\n");

  // write the variable definition
  fprintf(handle, "Binaries\n");
  cpt = 0;
  for (uint16 i = 0; i < pb->nb_genes; i++)
  {
    cpt += fprintf(handle, " Y%d", i);
    if (cpt > 80)
    {
      fprintf(handle, "\n");
      cpt = 0;
    }
  }
  for (uint16 j = 0; j < pb->nb_cond; j++)
  {
    cpt += fprintf(handle, " Z%d", j);
    if (cpt > 80)
    {
      fprintf(handle, "\n");
      cpt = 0;
    }
  }
  fprintf(handle, "\n");
  fprintf(handle, "End\n");

  // close the file
  fclose(handle);
}


/* ====================================================================== */
/* write the QP-model of the problem                                      */
/* ====================================================================== */

void writeQP (Problem * pb, const char name[])
{
  FILE * handle;
  uint16 cpt;

  // open the output file
  handle = fopen(name, "wt");

  // write the preamble

  // write the objective function
  fprintf(handle, "Minimize\ncost: [");
  cpt = 0;
  for (uint16 i = 0; i < pb->nb_genes; i++)
  {
    for (uint16 j = 0; j < pb->nb_cond; j++)
    {
      cpt += fprintf(handle, " %+d Y%d * Z%d", 2 * pb->C[i][j], i, j);
      if (cpt > 80)
      {
        fprintf(handle, "\n");
        cpt = 0;
      }
    }
  }
  fprintf(handle, "] / 2\n");

  // write the constraints
  fprintf(handle, "\nSubject to\n");
  fprintf(handle, "\n");

  // write the variable definition
  fprintf(handle, "Binaries\n");
  cpt = 0;
  for (uint16 i = 0; i < pb->nb_genes; i++)
  {
    cpt += fprintf(handle, " Y%d", i);
    if (cpt > 80)
    {
      fprintf(handle, "\n");
      cpt = 0;
    }
  }
  for (uint16 j = 0; j < pb->nb_cond; j++)
  {
    cpt += fprintf(handle, " Z%d", j);
    if (cpt > 80)
    {
      fprintf(handle, "\n");
      cpt = 0;
    }
  }
  fprintf(handle, "\n");
  fprintf(handle, "End\n");

  // close the file
  fclose(handle);
}
