#include "defines.h"

#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <mpi.h>
#include <time.h>


/* ====================================================================== */
/* chromosome memory allocation                                           */
/* ====================================================================== */

void allocChromosome (Chromosome * ch, uint16 size)
{
  if (ch == NULL)
    ERROR("invalid pointer");

  if (size == 0)
    ERROR("invalid size");

  ch->size = size;
  ch->conds = ALLOCATION(size,bool);
  ch->nb_active_g = 0;
  ch->nb_active_c = 0;
}


/* ====================================================================== */
/* chromosome memory cleaning                                             */
/* ====================================================================== */

void freeChromosome (Chromosome * ch)
{
  if (ch == NULL)
    ERROR("invalid pointer");

  FREE(ch->conds);
}


/* ====================================================================== */
/* random chromosome generation                                           */
/* ====================================================================== */

void randomChromosome (Chromosome * ch)
{
  if (ch == NULL)
    ERROR("invalid pointer");

  ch->nb_active_c = 0;
  for (uint16 i = 0; i < ch->size; i++)
  {
    bool used = Uniform (0, 1);
    ch->conds[i] = used;
	if(used) ++(ch->nb_active_c);
  }
}


/* ====================================================================== */
/* chromosome evaluation                                                  */
/* ====================================================================== */

void evalChromosome (Problem * pb, Chromosome * ch)
{
  int value, solutionGenes, val[pb->nb_genes], card = 0;

  // initialization
  for (uint16 i = 0; i < pb->nb_genes; i++)
    val[i] = 0;
  
  // cost computation: uses gene to reduce computation
  for (uint16 j = 0; j < pb->nb_cond; j++)
    if (ch->conds[j] == true)
    {
	  ++card;
      for (uint16 i = 0; i < pb->nb_genes; i++)
        val[i] += pb->C[i][j];
    }

  // final computation: only the negative-cost conditions
  value = 0; solutionGenes = 0;
  for (uint16 i = 0; i < pb->nb_genes; i++)
    if (val[i] < 0)
	{
      value += val[i];
	  ++solutionGenes;
    }
  ch->value = value;
  ch->nb_active_g = solutionGenes;
  ch->nb_active_c = card;
}


/* ====================================================================== */
/* display the chromosome                                                 */
/* ====================================================================== */

void displayChromosome (Chromosome * ch)
{
  printf("chromo: [ ");
  for (uint16 i = 0; i < ch->size; i++)
    printf("%2d ", ch->conds[i]);
  printf("] -> %6d (with %5d genes)\n", ch->value, ch->nb_active_g);
}

/* ====================================================================== */
/* sorther display of the chromosome                                      */
/* ====================================================================== */

void displayChromosomeShort (Chromosome * ch)
{
  printf("chromo: ");
  printf("%2d conds, %5d genes -> %6d\n", ch->nb_active_c, ch->nb_active_g, ch->value);
}


/* ====================================================================== */
/* detailed display of the chromosome                                     */
/* ====================================================================== */

void displayChromosomeLong (Chromosome * ch, Problem * pb)
{
  int value, solutionGenes, val[pb->nb_genes], card = 0;
  printf("chromo: \n\tconditions: ");
  // initialization
  for (uint16 i = 0; i < pb->nb_genes; i++)
    val[i] = 0;
  
  // cost computation: uses gene to reduce computation
  for (uint16 j = 0; j < pb->nb_cond; j++)
    if (ch->conds[j] == true)
    {
	  ++card;
	  printf("%d ",j);
      for (uint16 i = 0; i < pb->nb_genes; i++)
        val[i] += pb->C[i][j];
    }

  printf("\n\tgenes: ");
  // final computation: only the negative-cost conditions
  value = 0; solutionGenes = 0;
  for (uint16 i = 0; i < pb->nb_genes; i++)
    if (val[i] < 0)
	{
      value += val[i];
	  ++solutionGenes;
	  printf("%d ",i);
    }
  ch->value = value;
  printf("\n\tvalue: %d\n",ch->value);
  ch->nb_active_g = solutionGenes;
  ch->nb_active_c = card;
}


/* ====================================================================== */
/* population memory allocation                                           */
/* ====================================================================== */

void allocPopulation (Chromosome * ch, uint16 populationSize, uint16 chromosomeSize)
{
  if (ch == NULL)
    ERROR("invalid pointer");

  if (populationSize == 0)
    ERROR("invalid population size");

  if (chromosomeSize == 0)
    ERROR("invalid chromosome size");
	
  for(uint16 i=0; i < populationSize; ++i)
    allocChromosome(&(ch[i]), chromosomeSize);
}


/* ====================================================================== */
/* population memory cleaning                                             */
/* ====================================================================== */

void freePopulation (Chromosome * ch, uint16 populationSize)
{  
  if (ch == NULL)
    ERROR("invalid pointer");

  for(uint16 i=0; i < populationSize; ++i)
    freeChromosome(&(ch[i]));
}


/* ====================================================================== */
/* simple-mutation operator                                               */
/* ====================================================================== */

void simpleMutation (Chromosome * ch)
{  
  if (ch == NULL)
    ERROR("invalid pointer");

  uint16 index = Uniform(0,ch->size);
  ch->conds[index] = 1 - ch->conds[index];
}

/* ====================================================================== */
/* single-point cross-over operator                                       */
/* ====================================================================== */

void crossOver1 (Chromosome * ch1, Chromosome * ch2)
{  
  if (ch1 == NULL || ch2 == NULL)
    ERROR("invalid pointer");

  uint16 start = Uniform(0,ch1->size);
  bool temp;
  for(uint16 i=start; i<ch1->size; ++i)
  {
    temp = ch1->conds[i];
	ch1->conds[i] = ch2->conds[i];
	ch2->conds[i] = temp;
  }
}

/* ====================================================================== */
/* two-point cross-over operator                                          */
/* ====================================================================== */

void crossOver2 (Chromosome * ch1, Chromosome * ch2)
{  
  if (ch1 == NULL || ch2 == NULL)
    ERROR("invalid pointer");

  uint16 start = Uniform(0,ch1->size);
  uint16 end = Uniform(0,ch1->size);
  if(start > end)
  {
    uint16 temp = start;
	start = end;
	end = temp;
  }
  bool temp;
  for(uint16 i=start; i<end; ++i)
  {
    temp = ch1->conds[i];
	ch1->conds[i] = ch2->conds[i];
	ch2->conds[i] = temp;
  }
}



/* ====================================================================== */
/* local search operator                                                  */
/* ====================================================================== */

void localSearch (Problem * pb, Chromosome * ch)
{
  if (ch == NULL)
    ERROR("invalid pointer");

  Solution sol;
  allocSolution(&sol, pb->nb_nodes);
  clearSolution(&sol);
  chromosomeToSolution(pb, ch, &sol);
  LS6(pb, &sol);
  solutionToChromosome(pb, ch, &sol);
  freeSolution(&sol);
}


/* ====================================================================== */
/* copy of a chromosome into a solution                                   */
/* ====================================================================== */

void chromosomeToSolution (Problem * pb, Chromosome * ch, Solution * sol)
{  
  int val[pb->nb_genes];
  evalChromosome(pb,ch);
  
  // initialization
  for (uint16 i = 0; i < pb->nb_genes; i++)
    val[i] = 0;
  
  // adding active conditions to solution and computing costs
  for (uint16 j = 0; j < pb->nb_cond; j++)
  {
    if (ch->conds[j] == true)
    {
	  sol->vect[pb->nb_genes+j] = true;
      for (uint16 i = 0; i < pb->nb_genes; i++)
        val[i] += pb->C[i][j];
    }
	else
	  sol->vect[pb->nb_genes+j] = false;
  }
  
  // adding only the negative-cost genes to the solution
  for (uint16 i = 0; i < pb->nb_genes; i++)
  {
    if (val[i] < 0)
	  sol->vect[i] = true;
	else
	  sol->vect[i] = false;
  }
  
  sol->value = ch->value;
  sol->nb_active_c = ch->nb_active_c;
  sol->nb_active_g = ch->nb_active_g;

}



/* ====================================================================== */
/* copy of a solution into a chromosome                                   */
/* ====================================================================== */

void solutionToChromosome (Problem * pb, Chromosome * ch, Solution * sol)
{
  for (uint16 i = 0; i < pb->nb_cond; i++)
    ch->conds[i] = sol->vect[pb->nb_genes+i];
}



/* ====================================================================== */
/* diversification operator                                               */
/* ====================================================================== */

void diversification (Chromosome * ch, Chromosome * population, uint16 populationSize)
{  
  if (ch == NULL)
    ERROR("invalid pointer");

  uint16 frequency[ch->size];
  for(uint16 i=0; i < ch->size; ++i)
  {
    frequency[i] = 0;
    for(uint16 j=0; j < populationSize; ++j)
	{
	  if(population[j].conds[i])
	    ++frequency[i];
	}
  }
  
  for(uint16  i=0; i < ch->size; ++i)
  {
    uint16 rand = Uniform(0,populationSize);
	if(rand >= frequency[i])
	  ch->conds[i] = true;
	else
	  ch->conds[i] = false;
  }
}


/* ====================================================================== */
/* copy a chromosome from one position to another                         */
/* ====================================================================== */

void copyChromosome (Chromosome * ch_source, Chromosome * ch_dest)
{  
  if (ch_source == NULL || ch_dest == NULL)
    ERROR("invalid pointer");

  if (ch_source->size == 0)
  {
    printf("\ERROR: invalid size in copyChromosome, (corrected)\n");
	ch_source->size = ch_dest->size;
  }


  for(uint16 i=0; i<ch_source->size; ++i)
	ch_dest->conds[i] = ch_source->conds[i]; // (bool) is added just in case

  ch_dest->size = ch_source->size;
  ch_dest->nb_active_c = ch_source->nb_active_c;
  ch_dest->nb_active_g = ch_source->nb_active_g;
  ch_dest->value = ch_source->value;
}


/* ====================================================================== */
/* sort population of chromosomes according to their value                */
/* (based on Combsort11)                                                  */
/* ====================================================================== */

void sortPopulation (Chromosome * population, uint16 populationSize)
{  
  if (population == NULL)
    ERROR("invalid pointer");

  uint16 gap = populationSize;
  uint16 i, swaps;
    
  do
  {
    if(gap > 1)
	{
      gap = gap / 1.3;
      if(gap == 10 || gap == 9) gap = 11;
	}

    i = 0;
	swaps = 0;
	
	while(i + gap < populationSize)
	{
	  if (population[i].value > population[i+gap].value)
	  {
	    swapChromosomes(population, i, i+gap);
		swaps = 1;
	   }
	   ++i;
     }
       
  } while(gap > 1 || swaps != 0);

}


/* ====================================================================== */
/* swap two chromosomes in population (used in sorting algorithm)         */
/* ====================================================================== */

void swapChromosomes (Chromosome * population, uint16 ch1, uint16 ch2)
{  
  if (population == NULL)
    ERROR("invalid pointer");

  uint16 tempSize;
  int tempValue;
  bool tempCond;
  
  tempSize = population[ch1].size;
  population[ch1].size = population[ch2].size;
  population[ch2].size = tempSize;
  
  tempValue = population[ch1].nb_active_g;
  population[ch1].nb_active_g = population[ch2].nb_active_g;
  population[ch2].nb_active_g = tempValue;

  tempValue = population[ch1].nb_active_c;
  population[ch1].nb_active_c = population[ch2].nb_active_c;
  population[ch2].nb_active_c = tempValue;

  tempValue = population[ch1].value;
  population[ch1].value = population[ch2].value;
  population[ch2].value = tempValue;
  
  for(uint16 i=0; i < tempSize; ++i)
  {
    tempCond = population[ch1].conds[i];
    population[ch1].conds[i] = population[ch2].conds[i];
    population[ch2].conds[i] = tempCond;
  }
}


/* ====================================================================== */
/* remove multiple occurences of the same chromosome                      */
/* ====================================================================== */

uint16 removeRedundancy (Chromosome * population, uint16 populationSize)
{  
  if (population == NULL)
    ERROR("invalid pointer");
	
  uint16 i = 0;
  uint16 currentPopulation = populationSize;
  bool mustBeRemoved;

  while(i < currentPopulation - 1)
  {
	mustBeRemoved = identicalChromosomes(&(population[i]), &(population[i+1]));
    if(mustBeRemoved == true)
	{
	  removeChromosome(population, currentPopulation, i+1);
	  --currentPopulation;
	}
	else
	  ++i;
  }
  return currentPopulation;
}

/* ====================================================================== */
/* check whether two chromosomes are identical                            */
/* ====================================================================== */

bool identicalChromosomes (Chromosome * ch1, Chromosome * ch2)
{
  if (ch1 == NULL || ch2 == NULL)
    ERROR("invalid pointer");

  if (ch1->size != ch2->size)
  {
    printf("\ERROR: invalid sizes %d / %d, (corrected)\n", ch1->size, ch2->size);
	ch2->size = ch1->size;
  }
	
	
  bool areIdentical = true;
  uint16 i = 0;
  while(i < ch1->size)
  {
    if(ch1->conds[i] != ch2->conds[i])
	{
	  areIdentical = false;
	  i += ch1->size; // in order to safely exit from the loop
	}
    ++i;
  }
  return areIdentical;
}


/* ====================================================================== */
/* remove a chromosome from the population                                */
/* ====================================================================== */

void removeChromosome (Chromosome * population, uint16 populationSize, uint16 position)
{  
  if (population == NULL)
    ERROR("invalid pointer");
	
  uint16 i = position;
  while(i < populationSize-1)
  {
    copyChromosome(&(population[i+1]), &(population[i]));
	++i;
  }
}

/* ====================================================================== */
/* creation of a random initial population                                */
/* ====================================================================== */

void initPopulation (Problem * pb, Chromosome * population, uint16 populationSize)
{
  for(uint16 i=0; i<populationSize-1; ++i)
  {
    randomChromosome(&(population[i]));
    evalChromosome(pb, &(population[i]));
    //displayChromosomeShort(&(population[i]));
  }
  diversification(&(population[populationSize-1]), population, populationSize-1);
  evalChromosome(pb, &(population[populationSize-1]));
  //displayChromosomeShort(&(population[populationSize-1]));
}



/* ====================================================================== */
/* creation of an orthogonal initial population                           */
/* ====================================================================== */

void initOrthoPopulation (Problem * pb, Chromosome * population, uint16 populationSize)
{
  for(uint16 i=0; i<populationSize; ++i)
  {
    for(uint16 j=0; j < pb->nb_cond; ++j)
	{
	  if(i == j) population[i].conds[j] = 1;
	  else population[i].conds[j] = 0;
	}
    evalChromosome(pb, &(population[i]));
    //displayChromosomeShort(&(population[i]));
  }
}


/* ====================================================================== */
/* creation of an orthogonal initial population and adds complete sol     */
/* ====================================================================== */

void initOrthoPopulation2 (Problem * pb, Chromosome * population, uint16 populationSize)
{
  for(uint16 i=0; i<populationSize; ++i)
  {
    for(uint16 j=0; j < pb->nb_cond; ++j)
	{
	  if(i == j) population[i].conds[j] = 1;
	  else population[i].conds[j] = 0;
	}
    evalChromosome(pb, &(population[i]));
    //displayChromosomeShort(&(population[i]));
  }
  for(uint16 j=0; j < pb->nb_cond; ++j)
    population[populationSize].conds[j] = 1;
  evalChromosome(pb, &(population[populationSize]));
}



/* ====================================================================== */
/* main function                                                          */
/* ====================================================================== */

void genetic (Problem * pb, uint16 populationSize, Solution *sol, bool save)
{
  int procNum, clusterSize;
  MPI_Comm_rank(MPI_COMM_WORLD,&procNum);
  MPI_Comm_size(MPI_COMM_WORLD,&clusterSize);


  if(procNum==0)
  {  
    char cmd_date[64] = "date \"+DATE: %m/%d/%y%nTIME: %H:%M:%S\"";
    system(cmd_date);
  }

  uint32 randomSeed = 2*(procNum+time(0));
  seedMT(randomSeed);

  uint16 maxPopulation = 2*populationSize;
  uint16 migrationSize = populationSize/20;
//  if(migrationSize <= 0) migrationSize = 1;
  uint16 currentPopulation = 0;
  uint16 iterMaxGlobal = 10;
  uint16 iterMaxLocal = 100;
  uint16 mutationProba, crossOver1Proba, crossOver2Proba, localSearchProba;
  Chromosome population[maxPopulation];
  allocPopulation(population, maxPopulation, pb->nb_cond);
  
  if(procNum == 0)
  {
	mutationProba = 400; crossOver1Proba = 295; crossOver2Proba = 295; localSearchProba = 10;
  }
  else
  {
	mutationProba = 400; crossOver1Proba = 295; crossOver2Proba = 295; localSearchProba = 10;
  }
  
  uint16 richBuffer[migrationSize*pb->nb_cond];
  uint16 poorBuffer[migrationSize*pb->nb_cond];

  uint16 richSend;
  if(procNum==0) richSend = clusterSize-1;
  else richSend = procNum-1;
  uint16 poorSend;
  if(procNum==clusterSize-1) poorSend = 0;
  else poorSend = procNum+1;
  uint16 richRecv = poorSend;
  uint16 poorRecv = richSend;
	
  MPI_Barrier(MPI_COMM_WORLD);
  
  Chromosome keepBest[pb->nb_cond+1];
  allocPopulation(keepBest, pb->nb_cond+1, pb->nb_cond);

  // Initialisation of best solutions
  for(uint16 i=0; i<pb->nb_cond+1; ++i)
    keepBest[i].value = 0;

  // Initialisation of population
  initOrthoPopulation2 (pb, population, populationSize);
  currentPopulation = populationSize+1;
  
  // Main part of algorithm starts here

  // Global loop
  for(uint16 k=0; k < iterMaxGlobal; ++k)
  {
    MPI_Barrier(MPI_COMM_WORLD);
    if(procNum==3) printf("Global loop %d\n", k);

    // Local loop
    for(uint16 iter=0; iter < iterMaxLocal; ++iter)
    {
	  //if(procNum==3) printf("Local loop %d\n", iter);
	  
      // expansion phase
      while(currentPopulation < maxPopulation-1)
  	  { 
	    // preparing two targets
	    uint16 target = Uniform(0,currentPopulation-1);
	    copyChromosome(&(population[target]),&(population[currentPopulation]));
	    ++currentPopulation;
	    target = Uniform(0,currentPopulation-2);
	    copyChromosome(&(population[target]),&(population[currentPopulation]));
	    ++currentPopulation;

        // creating new solutions
        uint16 rand = Uniform(0,1000);
        if(rand <= mutationProba)
        {
	      // performing a simple mutation
		  simpleMutation(&(population[currentPopulation-2]));
 	  	  evalChromosome(pb, &(population[currentPopulation-2]));

	      // performing another one
		  simpleMutation(&(population[currentPopulation-1]));
		  evalChromosome(pb, &(population[currentPopulation-1]));
		  continue;
	    }
	    if(rand <= mutationProba + crossOver1Proba)
	    {
	      // performing a single-point cross-over
		  crossOver1(&(population[currentPopulation-2]),&(population[currentPopulation-1]));
		  evalChromosome(pb, &(population[currentPopulation-2]));
		  evalChromosome(pb, &(population[currentPopulation-1]));
		  continue;
	    }
	    if(rand <= mutationProba + crossOver1Proba + crossOver2Proba)
	    {
	      // performing a two-point cross-over
		  crossOver2(&(population[currentPopulation-2]),&(population[currentPopulation-1]));
		  evalChromosome(pb, &(population[currentPopulation-2]));
		  evalChromosome(pb, &(population[currentPopulation-1]));
		  continue;
	    }
	    if(rand <= mutationProba + crossOver1Proba + crossOver2Proba + localSearchProba)
	    {
	      // performing a local search
		  localSearch(pb, &(population[currentPopulation-2]));
		  evalChromosome(pb, &(population[currentPopulation-2]));		  

	      // performing a second local search
		  localSearch(pb, &(population[currentPopulation-1]));
		  evalChromosome(pb, &(population[currentPopulation-1]));		  
		  continue;
	    }
      }

	  // sorting and removing redundancy
	  sortPopulation(population, currentPopulation);
	  currentPopulation = removeRedundancy(population, currentPopulation);
	  
	  // updating best solutions
	  for(uint16 i=0; i < currentPopulation; ++i)
	  {
	    if(population[i].value < keepBest[population[i].nb_active_c].value)
		{
	      copyChromosome(&(population[i]), &(keepBest[population[i].nb_active_c]));
	    }
	  }
	  
	  // selection phase
	  while(currentPopulation > 0.9*populationSize)
	  {
	    // top 10% are conserved, tournament for the rest
		uint16 index1, index2, loser;
		index1 = Uniform(0.1*populationSize, currentPopulation-1);
		index2 = Uniform(0.1*populationSize, currentPopulation-1);
		if(population[index1].value > population[index2].value)
		  loser = index1;
		else
		  loser = index2;
		if(Uniform(0,1000) <= 300)
		  loser = index1 + index2 - loser;
		swapChromosomes(population, loser, currentPopulation-1);
		--currentPopulation;
	  }
	  
	  // diversification
	  while(currentPopulation < populationSize)
	  {
        diversification(&(population[currentPopulation]), population, currentPopulation);
	    ++currentPopulation;
	  }

    }

    // Local loop is over, migration is starting
	sortPopulation(population, currentPopulation);

	for(uint16 i=0; i < migrationSize; ++i)
	{
	  uint16 index = Uniform(0, currentPopulation/10);
	  for(uint16 j=0; j < pb->nb_cond; ++j)
	  {
	    if(population[index].conds[j] == true)
	      richBuffer[i*pb->nb_cond+j] = 1;
		else
		  richBuffer[i*pb->nb_cond+j] = 0;
	  }
	  removeChromosome(population, currentPopulation, index);
	  --currentPopulation;
	  
	  index = 9*currentPopulation/10 + Uniform(0, currentPopulation/10);
	  for(uint16 j=0; j < pb->nb_cond; ++j)
	  {
	    if(population[index].conds[j] == true)
		  poorBuffer[i*pb->nb_cond+j] = 1;
		else
		  poorBuffer[i*pb->nb_cond+j] = 0;
	  }
	  removeChromosome(population, currentPopulation, index);
	  --currentPopulation;
	}
	
	MPI_Send(richBuffer, migrationSize*pb->nb_cond, MPI_INT, richSend, k, MPI_COMM_WORLD);
	MPI_Send(poorBuffer, migrationSize*pb->nb_cond, MPI_INT, poorSend, k, MPI_COMM_WORLD);
	MPI_Recv(richBuffer, migrationSize*pb->nb_cond, MPI_INT, richRecv, k, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
	MPI_Recv(richBuffer, migrationSize*pb->nb_cond, MPI_INT, poorRecv, k, MPI_COMM_WORLD, MPI_STATUS_IGNORE);

	for(uint16 i=0; i < migrationSize; ++i)
	{
	  for(uint16 j=0; j < pb->nb_cond; ++j)
	  {
	    if(richBuffer[i*pb->nb_cond+j] == 1)
	      population[currentPopulation].conds[j] = true;
		else
		  population[currentPopulation].conds[j] = false;
	  }
	  population[currentPopulation].size = pb->nb_cond;
	  evalChromosome (pb, &(population[currentPopulation]));
	  ++currentPopulation;

	  for(uint16 j=0; j < pb->nb_cond; ++j)
	  {
	    if(poorBuffer[i*pb->nb_cond+j] == 1)
	      population[currentPopulation].conds[j] = true;
		else
		  population[currentPopulation].conds[j] = false;
	  }
	  population[currentPopulation].size = pb->nb_cond;
	  evalChromosome (pb, &(population[currentPopulation]));
	  ++currentPopulation;
	}
	sortPopulation(population, currentPopulation);
  }

  // gathering results
  
  MPI_Barrier(MPI_COMM_WORLD);
  if(procNum!=0)
  {
    printf("Algorithm over on node %d. Sendind results.\n", procNum);
    uint16 bestResultsBuffer[pb->nb_cond*pb->nb_cond];
	for(uint16 i=0; i < pb->nb_cond; ++i)
	{
      for(uint16 j=0; j < pb->nb_cond; ++j)
	  {
	    if(keepBest[i].conds[j] == true)
	      bestResultsBuffer[i*pb->nb_cond+j] = 1;
		else
		  bestResultsBuffer[i*pb->nb_cond+j] = 0;
	  }
    }
	MPI_Send(bestResultsBuffer, pb->nb_cond*pb->nb_cond, MPI_INT, 0, 0, MPI_COMM_WORLD);
  
  }
  
  else
  {
    Chromosome tempCh;
    printf("Algorithm over on node %d. Receiving results from %d other nodes:\n", procNum, clusterSize-1);
    uint16 bestResultsBuffer[pb->nb_cond*pb->nb_cond];
    for(uint16 k=1; k < clusterSize; ++k)
	{
	  MPI_Recv(bestResultsBuffer, pb->nb_cond*pb->nb_cond, MPI_INT, k, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);

	  for(uint16 i=0; i < pb->nb_cond; ++i)
	  {
        allocChromosome(&tempCh, pb->nb_cond);
	    for(uint16 j=0; j < pb->nb_cond; ++j)
		{
		  if(bestResultsBuffer[i*pb->nb_cond+j] == 1)
		    tempCh.conds[j] = true;
		  else
		    tempCh.conds[j] = false;
		}
		evalChromosome(pb, &tempCh);
		if(tempCh.nb_active_c != i) printf("WHY?\n");
		if(tempCh.value < keepBest[tempCh.nb_active_c].value)
		{
	      copyChromosome(&tempCh, &(keepBest[tempCh.nb_active_c]));
	    }
		freeChromosome(&tempCh);
	  }
	}
	// keepBest on proc 0 now has the best solutions
  }
  
  
  // displaying results
  
  if(procNum==0)
  {
	
    // display best solution for each number of categories
    for(uint16 i=0; i<pb->nb_cond+1; ++i)
      displayChromosomeShort(&(keepBest[i]));

    // detailed display of best solution for each number of categories
    // (remove comments if required)
    //printf("\n\n\n ********** \n ********** \n ********** \n\n\n");
    //for(uint16 i=0; i<pb->nb_cond+1; ++i)
    //  displayChromosomeLong(&(keepBest[i]),pb);

    if (save)
    {
      int ymax, ymin;
      char cmd[64];
      const char datafile[] = "enum.dat";
      const char outputfile[] = "enum.eps";
      const char cmdfile[]  = "display.p";
      FILE * handle;
  
      // open the data file
      if ((handle = fopen(datafile, "wt")) == NULL)
        ERROR("cannot open file %s\n", datafile);
  
      // write the data
      fprintf(handle, "# nb_conditions / value\n\n");
      ymax = INT_MIN; ymin = INT_MAX;
      for (uint16 j = 1; j < pb->nb_cond + 1; j++)
      {
        fprintf(handle, "%d %d\n", j, keepBest[j].value);
        if (keepBest[j].value > ymax)
          ymax = keepBest[j].value;
        if (ymin > keepBest[j].value)
          ymin = keepBest[j].value;
      }
      fprintf(handle, "\n");
  
      // close the file
      fclose(handle);
  
      // write down the command file
	  if ((handle = fopen(cmdfile, "wt")) == NULL)
        ERROR("cannot open file %s\n", cmdfile);
  
      // write the gnuplot commands
      fprintf(handle, "set encoding iso_8859_1\n");
      fprintf(handle, "unset log\nunset label\n");
      fprintf(handle, "set title \"impact of nb active conditions (%s)\"\n",
        pb->name);
      fprintf(handle, "set xlabel \"nb conditions\"\n");
      fprintf(handle, "set ylabel \"optimal value\"\n");
      fprintf(handle, "set xrange [%g:%g]\n", 0.0, pb->nb_cond + 0.5);
      for (uint16 j = 1; j < pb->nb_cond + 1; j++)
      {
        fprintf(handle, "set label \'%d\' at %g,%g\n",
	  	  keepBest[j].nb_active_g, 0.1 + (float)(j), (float)((ymin - ymax)/40 + keepBest[j].value));
      }
      fprintf(handle, "plot \'%s\' using 1:2 notitle with linespoints\n",
        datafile);
      fprintf(handle, "set term postscript eps\n");
      fprintf(handle, "set output \'%s\'\n", outputfile);
      fprintf(handle, "replot\n");
      fprintf(handle, "\n");
  
      // close the file
      fclose(handle);
      
      // run gnuplot
      sprintf(cmd, "gnuplot %s", cmdfile);
      system(cmd);
    }
  }

  MPI_Barrier(MPI_COMM_WORLD);
  freePopulation(population, maxPopulation);
  if(procNum==0)
  {
    char cmd_date[64] = "date \"+DATE: %m/%d/%y%nTIME: %H:%M:%S\"";
	system(cmd_date);
  }
}
