/* ======================================================================= */
/* paths.c:                                                                */
/*   functions to manage the paths for each commodity                      */
/*   two structures: normal list (Path) for easy storage                   */
/*                   dictionnary (PNode) for quick search                  */
/*                                                                         */
/*   structure consistency check has to be done!                           */
/*                                                                         */
/* created: 14/02/02 by Christophe Duhamel for the ILSP                    */
/* last modif: 12/02/07 by Christophe Duhamel                              */
/* ======================================================================= */

#include <string.h>

#include "prototypes.h"


/* ======================================================================= */
/* global variables to speed up the computation (avoid multiple alloc)     */
/* ======================================================================= */

static uint16 MAX_LENGTH = 0;


/* ======================================================================= */
/* creation of the structure that stores the paths                         */
/* ======================================================================= */

void CreateTableChemin (Commo * d, uint16 max_paths, uint16 max_length)
{
  Path   *chptr;
  PTable *tblptr;
  register uint16 i;

  /* memory allocation */
  tblptr = &(d->table);
  tblptr->max_paths = max_paths;
  tblptr->nb_paths = 0;
  tblptr->premier = NULL;
  tblptr->paths = chptr = ALLOCATION(max_paths, Path);

  /* paths initialisation */
  if (MAX_LENGTH < max_length)
    MAX_LENGTH = max_length;
  for (i = 0; i < max_paths; i++, chptr++)
  {
     chptr->id = i;
     chptr->od = d;
     chptr->nb_arcs = 0;
     chptr->arcs = ALLOCATION(max_length, Arc*);
  }
}


/* ======================================================================== */
/* destruction of the structure that stores the paths                       */
/* ======================================================================== */

void DestroyTableChemin (Commo * d)
{
  PNode  **pile, **top, *parcours, *pere;
  Path   *chptr;
  PTable *tblptr;
  register uint16 i;

  /* allocation of the stack used in the dico search/removal */
  tblptr = &(d->table);
  pile = ALLOCATION(tblptr->max_paths, PNode*);

  /* path tabel removal */
  for (i = 0, chptr = tblptr->paths; i < tblptr->max_paths; i++, chptr++)
    FREE(chptr->arcs);
  FREE(tblptr->paths);

  /* dico removal: tree search */
  top = pile;
  if (tblptr->premier != NULL)
  {
    *top = tblptr->premier;
    top++;
  }

  while (top != pile)
  {
    /* pop the forst node */
    top--;
    parcours = *top;

    /* go straight to a leaf */
    while (parcours != NULL)
    {
      /* go down one level */
      pere = parcours;
      parcours = parcours->fils;

      /* stroe the uncle */
      if (pere->frere != NULL)
      {
        *top = pere->frere;
        top++;
      }

      /* remove the father */
      FREE(pere);
    }
  }

  /* remove the search stack */
  FREE(pile);
}


/* ======================================================================== */
/* path table consistency check                                             */
/* ======================================================================== */

void CheckTableChemin (Commo * d)
{
  PNode  **pile, **top, *parcours, *pere;
  Arc    **path;
  uint16 *level, *lvlptr, lvl, cpt_ch;
  PTable *tblptr;
  Path   *chptr;
  register uint16 i, j;

#ifndef __DEBUG__
  return;
#endif

  IN_FCT();

  /* trivial checks */
  tblptr = &(d->table);
  if ((tblptr->paths == NULL) ||
      ((tblptr->nb_paths > 0) && (tblptr->premier == NULL)))
    ERROR("NULL ptr for commo %u", d->id_offic);

  /* stack allocation for the dico search */
  pile = ALLOCATION(tblptr->max_paths, PNode*);
  level = ALLOCATION(tblptr->max_paths, uint16);
  path = ALLOCATION(MAX_LENGTH, Arc*);

  /* tree search */
  top = pile;
  if (tblptr->premier != NULL)
  {
    *top = tblptr->premier;
    top++;
    lvlptr = level;
    *lvlptr = 0;
  }
  cpt_ch = 0;

  while (top != pile)
  {
    /* pop the first node */
    top--;
    parcours = *top;
    lvl = *lvlptr;
    lvlptr--;
    path[lvl] = parcours->arc;

    /* do down straight to a leaf */
    while (parcours != NULL)
    {
      /* go down one level */
      pere = parcours;
      parcours = parcours->fils;
      if (parcours != NULL)
      {
        path[lvl+1] = parcours->arc;
      }

      /* store the uncle */
      if (pere->frere != NULL)
      {
        *top = pere->frere;
        top++;
        lvlptr++;
        *lvlptr = lvl;

        if (pere->frere->arc->id <= pere->arc->id)
          ERROR("violated lexicographic order (%d,%d) for commo %u",
            pere->arc->id, pere->frere->arc->id, d->id_offic);
      }
      lvl++;
    }

    if (pere->path == NULL)
      ERROR("final node does not link to path for commo %u", d->id_offic);
    chptr=pere->path;
    if (lvl != chptr->nb_arcs)
      ERROR("unmatched dico and table path size for commo %u", d->id_offic);
    for (i=0; i<chptr->nb_arcs; i++)
      if (chptr->arcs[i] != path[i])
        ERROR("unmatched dico and table path for commo %u", d->id_offic);
    cpt_ch++;
  }

  if (cpt_ch != tblptr->nb_paths)
    ERROR("unmatched number of dico and table paths for commo %u", d->id_offic);

  for (i = 0, cpt_ch = 0; i < tblptr->nb_paths; i++)
  {
    chptr = tblptr->paths + i;
    if (chptr->od != d)
      ERROR("path without a commodity for commo %u", d->id_offic);
    if (chptr->arcs == NULL)
      ERROR("path without an arc list for commo %u", d->id_offic);
    if ((chptr->actif == true) && (chptr->wave != UNASSIGNED))
    {
      cpt_ch++;
      for (j = 0; (j < d->nb_actifs) && (d->actifs[j] != chptr); j++) ;
      if (j == d->nb_actifs)
        ERROR("unreferenced active path for commo %u", d->id_offic);
    }
  }
  if (d->nb_actifs != cpt_ch)
    ERROR("unmatched number of active paths for commo %u", d->id_offic);

  /* stack removal */
  FREE(pile);
  FREE(level);
  FREE(path);

  OUT_FCT();
}


/* ======================================================================= */
/* displays a path                                                         */
/* ======================================================================= */

void DisplayChemin (Path * ch)
{
  register uint16 i;

  IN_FCT();

  printf("path %3d from commodity %3d (%3d -> %3d) : wavelength %d on [%2d",
    ch->id, ch->od->id_offic, ch->od->orig->id_offic, ch->od->dest->id_offic, ch->wave,
    ch->od->orig->id_offic);

  /* display the nodes on the path */
  for (i = 0; i < ch->nb_arcs; i++)
    printf(" %2d", ch->arcs[i]->dest->id);
  printf("]\n");
  fflush(stdout);

  OUT_FCT();
}


/* ======================================================================= */
/* briefly displays the path tables                                        */
/* ======================================================================= */

void DisplayTableChemin (Commo * d)
{
  Arc    **arcptr;
  Path   *chptr;
  PTable *tblptr;
  register uint16 i, j;

  IN_FCT();

  /* for all the paths */
  tblptr = &(d->table);
  printf(" commodity %3d (%2d -> %2d) %2d paths [used: %2d / %2d] ->",
    d->id_offic, d->orig->id_offic, d->dest->id_offic, d->demand, d->nb_actifs, tblptr->nb_paths);
  for (i = 0; i < d->nb_actifs; i++)
    printf("%2d", d->actifs[i]->id);
  printf(")\n");
  for (i = 0, chptr = tblptr->paths; i < tblptr->nb_paths; i++, chptr++)
  {
    printf("\tpath %3d (com %3d) : %s wavelength = %2d on [%2d",
      chptr->id, chptr->od->id, ((chptr->actif==true) ? "active" : "  idle"),
      chptr->wave, chptr->od->orig->id_offic);

    arcptr = chptr->arcs;
    for (j = 0; j < chptr->nb_arcs; j++, arcptr++)
      printf(" %2d", (*arcptr)->dest->id_offic);
    printf("]\n");
    /* DisplayChemin (chptr); */
  }
  fflush(stdout);

  OUT_FCT();
}


/* ======================================================================= */
/* displays the path dictionnary                                           */
/* ======================================================================= */

void DisplayDicoChemin (Commo * d)
{
  PNode  **pile, **top, *parcours, *pere;
  uint16 *level, *lvlptr, lvl;
  PTable *tblptr;
  register uint16 i;

  IN_FCT();

  /* allocation for the dico search stack */
  tblptr = &(d->table);
  pile = ALLOCATION(tblptr->max_paths, PNode*);
  level = ALLOCATION(tblptr->max_paths, uint16);

  /* tree search */
  top = pile;
  if (tblptr->premier != NULL)
  {
    *top = tblptr->premier;
    top++;
    lvlptr = level;
    *lvlptr = 0;
  }

  while (top != pile)
  {
    /* pop the first node */
    top--;
    parcours = *top;
    lvl = *lvlptr;
    lvlptr--;
    for (i = 0; i < lvl; i++)
      printf("          ");
    printf("(%3d->%3d)", parcours->arc->orig->id_offic, parcours->arc->dest->id_offic);
    fflush(stdout);

    /* go down straight to a leaf */
    while (parcours != NULL)
    {
      /* go down one level */
      pere = parcours;
      parcours = parcours->fils;
      if (parcours != NULL)
      {
        printf("(%3d->%3d)", parcours->arc->orig->id_offic, parcours->arc->dest->id_offic);
        fflush(stdout);
      }

      /* store the uncle */
      if (pere->frere != NULL)
      {
        *top =pere->frere;
        top++;
        lvlptr++;
        *lvlptr = lvl;
      }
      lvl++;
    }
    printf("\n");
    fflush(stdout);
  }

  /* remove the stack */
  FREE(pile);
  FREE(level);

  OUT_FCT();
}


/* ======================================================================= */
/* insert one path into the associated path table                          */
/* ======================================================================= */

Path * InsertTableChemin (Commo * d, Path * ch)
{
  PNode  *courant, *pere, *frere;
  PTable *tblptr;
  Path   *pathptr, **path2ptr, **chptr;
  Arc    **arcptr, *aptr;
  register uint16 i, j, k;

  IN_FCT();

  /* initialisation */
  tblptr = &(d->table);

  /* skip over the shared part */
  pere = frere = NULL;
  courant = tblptr->premier;
  for (i = 0, arcptr = ch->arcs; (courant != NULL) && (i < ch->nb_arcs); i++, arcptr++)
  {
    /* look for the right node */
    while ((courant != NULL) && (courant->arc->id < (*arcptr)->id))
    {
      frere = courant;
      courant = courant->frere;
    }

    /* check if this node has been found */
    if ((courant == NULL) || (courant->arc->id > (*arcptr)->id))
      break;

    /* go down to the first child */
    pere = courant;
    frere = NULL;
    courant =  courant->fils;
  }

  /* existing path: only update infos */
  if (i == ch->nb_arcs)
  {
    pathptr = pere->path;

    if (pathptr->actif == ch->actif)
      return (pathptr);

    if (pathptr->actif == false)
    {
      /* path activation */
      pathptr->wave = ch->wave;
      d->actifs[d->nb_actifs] = pathptr;
      pathptr->actif = true;
      d->nb_actifs++;
      for (j = 0, arcptr = pathptr->arcs; j < pathptr->nb_arcs; j++, arcptr++)
      {
        aptr = *arcptr;
        aptr->paths[aptr->nb_paths] = pathptr;
        aptr->nb_paths++;
      }
    }
    else
    {
      /* path desactivation */
      /* remove ch from the active paths list */
      for (j = 0, path2ptr = d->actifs; (*path2ptr) != ch; j++, path2ptr++) ;
      for (j++, path2ptr++; j < d->nb_actifs; j++, path2ptr++)
        *(path2ptr - 1) = *path2ptr;
      d->nb_actifs--;
      pathptr = *path2ptr;
      pathptr->actif = false;
      pathptr->wave = UNASSIGNED;

      /* update the arcs in ch */
      for (j = 0, arcptr = pathptr->arcs; j < pathptr->nb_arcs; j++, arcptr++)
      {
        /* remove it from arcptr path table */
        aptr = *arcptr;
        for (k = 0, chptr = aptr->paths; (*chptr) != ch; j++, chptr++) ;
        for (k++, chptr++; k < aptr->nb_paths; k++, chptr++)
          *(chptr - 1) = *chptr;
        aptr->nb_paths--;
      }
    }
    return (pathptr);
  }

  /* security check */
  if (tblptr->nb_paths == tblptr->max_paths)
  {
    ERROR("no more room");
    /* return (NULL); */
  }

  /* insert the missing part: first node */
  courant = ALLOCATION(1, PNode);
  courant->arc = *arcptr;
  if (frere != NULL)
  {
    courant->frere = frere->frere;
    frere->frere = courant;
  }
  else
  {
    if (pere == NULL)
    {
      courant->frere = tblptr->premier;
      tblptr->premier = courant;
      /* courant->frere = NULL; */
    }
    else
    {
      courant->frere = pere->fils;
      pere->fils = courant;
    }
  }
  courant->path = NULL;
  pere = courant;

  /* insert the missing part: path tail */
  for (i++, arcptr++; i < ch->nb_arcs; i++, arcptr++)
  {
    courant = ALLOCATION(1, PNode);
    courant->arc = *arcptr;
    courant->frere = NULL;
    courant->path = NULL;
    pere->fils = courant;
    pere = courant;
  }
  pere->fils = NULL;
  pere->path = pathptr = tblptr->paths + tblptr->nb_paths;

  /* add the path into the table: o(m) */
  pathptr->id = tblptr->nb_paths;
  pathptr->nb_arcs = ch->nb_arcs;
  pathptr->wave = ch->wave;
  pathptr->actif = ch->actif;
  for (i = 0; i < ch->nb_arcs; i++)
    pathptr->arcs[i] = ch->arcs[i];
  tblptr->nb_paths++;
  if (pathptr->actif == true)
  {
    for (i = 0; i < ch->nb_arcs; i++)
    {
      /* add the path into the arc path table */
      aptr = ch->arcs[i];
      aptr->paths[aptr->nb_paths] = pathptr;
      aptr->nb_paths++;
    }
    d->actifs[d->nb_actifs] = pathptr;
    d->nb_actifs++;
  }

  OUT_FCT();

  return (pathptr);
}


/* ======================================================================= */
/* check if a path belongs to the associated table                         */
/* ======================================================================= */

bool IsInTableChemin (Commo * d, Path * ch)
{
  PNode  *nodeptr;
  PTable *tblptr;
  Arc    **arcptr;
  register uint16 i;

  IN_FCT();

  /* security check */
  tblptr = &(d->table);
  if (tblptr->nb_paths == 0)
    return (false);

  nodeptr = tblptr->premier;
  for (i = 0, arcptr = ch->arcs; i < ch->nb_arcs; i++, arcptr++)
  {
    /* look for the associated node */
    while ((nodeptr != NULL) && (nodeptr->arc->id < (*arcptr)->id))
      nodeptr = nodeptr->frere;
    if ((nodeptr == NULL) || (nodeptr->arc->id > (*arcptr)->id))
      return (false);
    nodeptr = nodeptr->fils;
  }
  if (nodeptr != NULL)
    ERROR("strange path");

  OUT_FCT();

  return (true);
}


/* ======================================================================= */
/* activate a path from the path table                                     */
/* ======================================================================= */

void ActivatePath (Path * p, uint16 l)
{
  Arc             *aptr;
  Commo           *dptr;
  register uint16 i;

  p->wave = l;
  p->actif = true;

  for (i = 0; i < p->nb_arcs; i++)
  {
    /* add the path into the arc path table */
    aptr = p->arcs[i];
    aptr->paths[aptr->nb_paths] = p;
    aptr->nb_paths++;
  }
  dptr = p->od;
  dptr->actifs[dptr->nb_actifs] = p;
  dptr->nb_actifs++;
}


/* ======================================================================= */
/* end of paths.c                                                          */
/* ======================================================================= */
