#include "defines.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>


const float FACTOR = 100.0;


/* ====================================================================== */
/* nearest int rounding                                                   */
/* ====================================================================== */

static double Nint (double r)
{
  double r1 = floor(r);

  return ((r - r1 <= 0.5) ? r1 : 1.0 + r1);
}


/* ====================================================================== */
/* problem memory allocation                                              */
/* ====================================================================== */

void allocProblem (Problem * pb, uint16 nb_g, uint16 nb_c)
{
  if (pb == NULL)
    ERROR("invalid pointer");

  if ((nb_g == 0) || (nb_c == 0))
    ERROR("invalid size");

  pb->nb_genes = nb_g;
  pb->nb_cond = nb_c;
  pb->nb_nodes = nb_g + nb_c;
  pb->C = ALLOCATION(nb_g,int*);
  *(pb->C) = ALLOCATION(nb_g*nb_c,int);
  for (uint16 i = 1; i < nb_g; i++)
    pb->C[i] = *(pb->C) + i * nb_c;
  pb->used = ALLOCATION(pb->nb_nodes,bool);
  pb->used_g = pb->used;
  pb->used_c = pb->used + nb_g;
  pb->cost = ALLOCATION(pb->nb_nodes,int);
  pb->cost_g = pb->cost;
  pb->cost_c = pb->cost + nb_g;
}


/* ====================================================================== */
/* problem memory cleaning                                                */
/* ====================================================================== */

void freeProblem (Problem * pb)
{
  if (pb == NULL)
    ERROR("invalid pointer");

  FREE(*(pb->C));
  FREE(pb->C);
  FREE(pb->used);
  FREE(pb->cost);
}


/* ====================================================================== */
/* compute the initial zero solution                                      */
/* ====================================================================== */

void clearGraphSolution (Problem * pb)
{
  // clear the node structure
  for (int i = 0; i < pb->nb_nodes; i++)
  {
    pb->used[i] = false;
    pb->cost[i] = 0;
  }

  // solution cost
  pb->value = 0;
  pb->nb_active_g = pb->nb_active_c = 0;
}


/* ====================================================================== */
/* random problem generation                                              */
/* ====================================================================== */

void randomProblem1 (Problem * pb)
{
  if (pb == NULL)
    ERROR("invalid pointer");

  if (pb->C == NULL)
    ERROR("cost matrix is not allocated");

  for (uint16 i = 0; i < pb->nb_genes; i++)
    for (uint16 j = 0; j < pb->nb_cond; j++)
      pb->C[i][j] = Uniform (-100, 100);

  sprintf(pb->name, "random_%d_%d.txt", pb->nb_genes, pb->nb_cond);
}


/* ====================================================================== */
/* random problem generation                                              */
/* MIT-style data: 10% neg in [-10,0], 90% pos in [0,60]                  */
/* ====================================================================== */

void randomProblem2 (Problem * pb)
{
  if (pb == NULL)
    ERROR("invalid pointer");

  if (pb->C == NULL)
    ERROR("cost matrix is not allocated");

  // data generation
  for (uint16 j = 0; j < pb->nb_cond; j++)
  {
    double proba_neg = randNorm (0.11, 0.006);
    for (uint16 i = 0; i < pb->nb_genes; i++)
    {
      if (randExc() < proba_neg)
      {
        int min = (int)(randNorm (-9.0, 0.2));
        pb->C[i][j] = Uniform (min, 0);
      }
      else
      {
        int max = (int)(randNorm (60.0, 1.3));
        pb->C[i][j] = Uniform (0, max);
      }
    }
  }

  sprintf(pb->name, "randomMIT_%d_%d.txt", pb->nb_genes, pb->nb_cond);
}


/* ====================================================================== */
/* read an instance according to Dimitri-format                           */
/* ====================================================================== */

void readProblem (Problem * pb, const char filename[])
{
  FILE * handle;
  uint16 nb_g, nb_c;

  if (pb == NULL)
    ERROR("invalid pointer");

  // open the data file
  handle = fopen(filename, "rt");
  if (handle == NULL)
    ERROR("cannot open file %s", filename);

  // set the problem's name
  strcpy(pb->name, filename);

  // read the data size
  fscanf(handle, "%d%d", &nb_g, &nb_c);

  // data structures initialization
  allocProblem (pb, nb_g, nb_c);

  // read the cost of the arcs
  for (uint16 i = 0; i < nb_g; i++)
    for (uint16 j = 0; j < nb_c; j++)
    {
      double d;
      fscanf(handle, "%lf", &d);
      pb->C[i][j] = Nint(d * FACTOR);
    }

  // close the file
  fclose(handle);
}


/* ====================================================================== */
/* save an instance according to Dimitri-format                           */
/* ====================================================================== */

void writeProblem (Problem * pb, const char filename[])
{
  FILE * handle;

  if (pb == NULL)
    ERROR("invalid pointer");

  // open the data file
  handle = fopen(filename, "wt");
  if (handle == NULL)
    ERROR("cannot open file %s", filename);

  // write the data size
  fprintf(handle, "%d %d\n", pb->nb_genes, pb->nb_cond);

  // write the cost of the arcs
  for (uint16 i = 0; i < pb->nb_genes; i++)
  {
    for (uint16 j = 0; j < pb->nb_cond; j++)
      fprintf(handle, " %g", (float)(pb->C[i][j]) / FACTOR);
    fprintf(handle, "\n");
  }

  // close the file
  fclose(handle);
}


/* ====================================================================== */
/* display the problem data                                               */
/* ====================================================================== */

void displayProblem (Problem * pb)
{
  printf("instance %s\n%4d genes, %2d conditions:\n",
    pb->name, pb->nb_genes, pb->nb_cond);
  for (uint16 i = 0; i < pb->nb_genes; i++)
  {
    for (uint16 j = 0; j < pb->nb_cond; j++)
      printf(" %4d", pb->C[i][j]);
    printf("\n");
  }
}


/* ====================================================================== */
/* display a solution to the problem                                      */
/* ====================================================================== */

void displayGraphSolution (Problem * pb, bool reduced)
{
  printf("solution: value = %d\n", pb->value);
  printf("%d genes:", pb->nb_active_g);
  for (uint16 i = 0; i < pb->nb_genes; i++)
  {
    if ((pb->used_g[i] == true) && (reduced == false))
      printf(" %d", i);
  }
  printf("\n%d conds:", pb->nb_active_c);
  for (uint16 j = 0; j < pb->nb_cond; j++)
  {
    if ((pb->used_c[j] == true) && (reduced == false))
      printf(" %d", j);
  }
  printf("\n");
}


/* ====================================================================== */
/* check a solution to the problem                                        */
/* ====================================================================== */

void checkGraphSolution (Problem * pb)
{
  int total;
  int contrib[pb->nb_nodes];
  uint16 nb_active_g, nb_active_c;

  // initialization
  for (uint16 i = 0; i < pb->nb_nodes; i++)
    contrib[i] = 0;

  // check for the cost
  nb_active_g = nb_active_c = 0;
  total = 0;
  for (uint16 i = 0; i < pb->nb_genes; i++)
  {
    if (pb->used_g[i] == true)
    {
      for (uint16 j = 0; j < pb->nb_cond; j++)
      {
        if (pb->used_c[j] == true)
        {
          total += pb->C[i][j];
          contrib[i] += pb->C[i][j];
          contrib[pb->nb_genes + j] += pb->C[i][j];
        }
      }
    }
  }
  if (total != pb->value)
    ERROR("wrong value (given %d, checked %d)", pb->value, total);


  // check for the card
  nb_active_g = 0;
  for (uint16 i = 0; i < pb->nb_genes; i++)
  {
    if (pb->used_g[i] == true)
      nb_active_g++;
  }
  if (nb_active_g != pb->nb_active_g)
    ERROR("wrong active genes (given %d, checked %d)",
      pb->nb_active_g, nb_active_g);
  nb_active_c = 0;
  for (uint16 j = 0; j < pb->nb_cond; j++)
  {
    if (pb->used_c[j] == true)
      nb_active_c++;
  }
  if (nb_active_c != pb->nb_active_c)
    ERROR("wrong active conditions (given %d, checked %d)",
      pb->nb_active_c, nb_active_c);

  // check for the contribs
/*
  for (uint16 i = 0; i < pb->nb_nodes; i++)
    if ((pb->used[i] == true) && (contrib[i] >= 0))
    {
      displayGraphSolution (pb, false);
      ERROR("element %d with positive (%d) contribution", i, contrib[i]);
    }
*/
}


/* ====================================================================== */
/* preprocessing: remove any gene that do not react well                  */
/* could try on the conditions too                                        */
/* ====================================================================== */

void reduceProblem (Problem * pb)
{
  bool bad;
  uint16 cpt, k;
  int ** new_C, * new_cost;
  bool * new_used;

  if (pb == NULL)
    ERROR("invalid pointer");

  // check for the total cost
  cpt = 0;
  for (uint16 i = 0; i < pb->nb_genes; i++)
  {
    // initialisation
    bad = false;

    // check if the gene is OK
    for (uint16 j = 0; j < pb->nb_cond; j++)
    {
      if (pb->C[i][j] < 0)
      {
        bad = true;
        break;
      }
    }

    // update
    pb->used_g[i] = bad;
    if (bad == false)
      cpt++;
  }
  cpt = pb->nb_genes - cpt;

  // allocation of the new structures
  new_C = ALLOCATION(cpt,int*);
  *(new_C) = ALLOCATION(cpt*pb->nb_cond,int);
  for (uint16 i = 1; i < cpt; i++)
    new_C[i] = *(new_C) + i * pb->nb_cond;
  new_used = ALLOCATION(cpt + pb->nb_cond,bool);
  new_cost = ALLOCATION(cpt + pb->nb_cond,int);

  // transfer the costs
  k = 0;
  for (uint16 i = 0; i < pb->nb_genes; i++)
  {
    if (pb->used_g[i] == false)
      continue;

    for (uint16 j = 0; j < pb->nb_cond; j++)
      new_C[k][j] = pb->C[i][j];
    k++;
  }

  // delete the old data
  FREE(*(pb->C));
  FREE(pb->C);
  FREE(pb->used);
  FREE(pb->cost);

  // set the new data
  pb->nb_genes = cpt;
  pb->nb_nodes = cpt + pb->nb_cond;
  pb->C = new_C;
  pb->used = new_used;
  pb->used_g = pb->used;
  pb->used_c = pb->used + cpt;
  pb->cost = new_cost;
  pb->cost_g = pb->cost;
  pb->cost_c = pb->cost + cpt;

  // check the all-neg genes and cond
  bool neg_g[pb->nb_genes], neg_c[pb->nb_cond];
  for (uint16 i = 0; i < pb->nb_genes; i++)
    neg_g[i] = true;
  for (uint16 j = 0; j < pb->nb_cond; j++)
    neg_c[j] = true;
  for (uint16 i = 0; i < pb->nb_genes; i++)
  {
    for (uint16 j = 0; j < pb->nb_cond; j++)
    {
      if (pb->C[i][j] > 0)
      {
        neg_g[i] = false;
        neg_c[j] = false;
      }
    }
  }
  uint16 cpt_g = 0;
  for (uint16 i = 0; i < pb->nb_genes; i++)
    if (neg_g[i] == true)
      cpt_g++;
  uint16 cpt_c = 0;
  for (uint16 j = 0; j < pb->nb_cond; j++)
    if (neg_c[j] == true)
      cpt_c++;
  printf("reduction test: %d gene, %d cond all negative\n", cpt_g, cpt_c);
}
