#include "defines.h"

#include <stdlib.h>
#include <stdio.h>


/* ====================================================================== */
/* solution memory allocation                                             */
/* ====================================================================== */

void allocSolution (Solution * s, uint16 size)
{
  if (s == NULL)
    ERROR("invalid pointer");

  if (size == 0)
    ERROR("invalid size");

  s->size = size;
  s->vect = ALLOCATION(size,bool);
}


/* ====================================================================== */
/* solution memory cleaning                                               */
/* ====================================================================== */

void freeSolution (Solution * s)
{
  if (s == NULL)
    ERROR("invalid pointer");

  FREE(s->vect);
}


/* ====================================================================== */
/* compute the initial zero solution                                      */
/* ====================================================================== */

void clearSolution (Solution * s)
{
  // init the vector
  for (uint16 i = 0; i < s->size; i++)
    s->vect[i] = false;

  // solution cost
  s->value = 0;
  s->nb_active_c = s->nb_active_g = 0;
}


/* ====================================================================== */
/* compute the initial zero solution                                      */
/* ====================================================================== */

void copySolution (Solution * s1, Solution * s2)
{
  // safety check
  if (s1->size != s2->size)
    ERROR("mismatching sizes");

  // copy the vector
  for (uint16 i = 0; i < s1->size; i++)
    s2->vect[i] = s1->vect[i];

  // solution cost
  s2->value = s1->value;
  s2->nb_active_c = s1->nb_active_c;
  s2->nb_active_g = s1->nb_active_g;
}


/* ====================================================================== */
/* random solution generation                                             */
/* ====================================================================== */

void randomSolution (Solution * s, Problem * pb)
{
  if (s == NULL)
    ERROR("invalid pointer");

  // set the state vector
  for (uint16 i = 0; i < s->size; i++)
    s->vect[i] = Uniform (0, 1);

  // compute its value: o(m)
  int value = 0;
  for (uint16 i = 0; i < pb->nb_genes; i++)
    if (s->vect[i] == true)
    {
      for (uint16 j = 0; j < pb->nb_cond; j++)
        if (s->vect[j + pb->nb_genes] == true)
          value += pb->C[i][j];
    }
  s->value = value;

  // compute its cardinality: 0(n+m)
  s->nb_active_g = 0;
  for (uint16 i = 0; i < pb->nb_genes; i++)
    if (s->vect[i] == true)
      s->nb_active_g++;
  s->nb_active_c = 0;
  for (uint16 j = 0; j < pb->nb_cond; j++)
    if (s->vect[j + pb->nb_genes] == true)
      s->nb_active_c++;
}


/* ====================================================================== */
/* display the solution                                                   */
/* ====================================================================== */

void displaySolution (Solution * s, bool detailled)
{
  printf("solution: [");
  if (detailled == true)
  {
    for (uint16 i = 0; i < s->size; i++)
      printf("%2d", s->vect[i]);
  }
  else
  {
    printf("%d genes, %d conds", s->nb_active_g, s->nb_active_c);
  }
  printf("] -> %6d\n", s->value);
}


/* ====================================================================== */
/* display the solution                                                   */
/* ====================================================================== */

void saveSolution (Solution * s, const char filename[])
{
  FILE * handle;
  uint16 cpt;

  // open the file
  handle = fopen(filename, "wt");
  if (handle == NULL)
    ERROR("unable to save to file %s", filename);

  // write the size and the value
  fprintf(handle, "%d %d %d %d\n",
    s->size, s->value, s->nb_active_g, s->nb_active_c);

  // write the decision vector
  cpt = 0;
  for (uint16 i = 0; i < s->size; i++)
  {
    cpt += fprintf(handle, "%2d", s->vect[i]);
    if (cpt > 80)
    {
      fprintf(handle, "\n");
      cpt = 0;
    }
  }
  fprintf(handle, "\n");

  // close the file
  fclose(handle);
}


/* ====================================================================== */
/* display the solution                                                   */
/* ====================================================================== */

void extractSolution (Problem * pb, Solution * s)
{
  if ((pb == NULL) || (s == NULL))
    ERROR("invalid pointer");

  if (pb->nb_nodes != s->size)
    ERROR("mismatching sizes");

  // copy the state vectors
  for (uint16 i = 0; i < pb->nb_nodes; i++)
    s->vect[i] = pb->used[i];

  // copy the value
  s->value = pb->value;
  s->nb_active_g = pb->nb_active_g;
  s->nb_active_c = pb->nb_active_c;
}


/* ====================================================================== */
/* insert the solution into the problem structure                         */
/* ====================================================================== */

void insertSolution (Problem * pb, Solution * s)
{
  if ((pb == NULL) || (s == NULL))
    ERROR("invalid pointer");

  if (pb->nb_genes + pb->nb_cond != s->size)
    ERROR("mismatching sizes");

  // copy the state vectors
  for (uint16 i = 0; i < pb->nb_nodes; i++)
  {
    pb->used[i] = s->vect[i];
    pb->cost[i] = 0;
  }

  // copy the value
  pb->value = s->value;
  pb->nb_active_g = s->nb_active_g;
  pb->nb_active_c = s->nb_active_c;

  // set up the add/drop costs
  for (uint16 i = 0; i < pb->nb_genes; i++)
  {
    bool cond_g = pb->used_g[i];

    for (uint16 j = 0; j < pb->nb_cond; j++)
    {
      int tmp = pb->C[i][j];

      if (cond_g == true)
      {
        if (pb->used_c[j] == true)
        {
          pb->cost_g[i] -= tmp;
          pb->cost_c[j] -= tmp;
        }
        else
          pb->cost_c[j] += tmp;
      }
      else
      {
        if (pb->used_c[j] == true)
          pb->cost_g[i] += tmp;
      }
    }
  }
}
