#include <fstream>
#include <strstream>
#include <cmath>
#include <string>
#include "matrice.h"

#define WAIT_MSG "Enter 0 to go back to the main menu..."



/************************/
/* Creating the weights */
/************************/

int create_weights(string microarrayFile, int nbBins) {
  ifstream inFile;
  inFile.open(microarrayFile.c_str(), ios::in);
  int lines, columns;

  if (inFile.bad())
    return (1); // Error while opening... Program stops.
 
  // reads the file to get the number of lines (i.e. genes) and columns (i.e. conditions)
  inFile >> lines >> columns;

  // allocates matrices  
  Matrice data(lines,columns,0);  
  Matrice norm(lines,columns,0);  
  Matrice categ(lines,columns,0);
  Matrice weights(lines,columns,0);

  // allocates arrays
  double* total_J = new double[lines];
  double* total_I = new double[columns];
  double* min_J = new double[columns];
  double* max_J = new double[columns];


  for(int i=0;i<lines;++i) total_J[i]=0;
  for(int i=0;i<columns;++i) { min_J[i]=10000; max_J[i]=0; total_I[i]=0; }

  // reads the data to create the matrix and save the ums over lines and columns
  for(int i=0;i<lines;++i) {
          for(int j=0;j<columns;++j) {
                  inFile >> data(i,j);   // value added to the data matrix
                  total_J[i]+=data(i,j); // the total value of line i over all columns is updated
                  total_I[j]+=data(i,j); // the total value of column j over all lines is updated
          }
  }
  inFile.close();  

  // creates the matrix containing the expression ratio
  for(int i=0;i<lines;++i) {
          double average=total_J[i]/columns;
          for(int j=0;j<columns;++j) {
                  norm(i,j)=data(i,j)/average;
                  if(min_J[j]>norm(i,j) && norm(i,j)!=0) min_J[j]=norm(i,j); // the min value of column j is updated
                  if(max_J[j]<norm(i,j)) max_J[j]=norm(i,j); // the max value of column j is updated
          }
  }

  // for each condition, we sort the values into bins
  for(int j=0;j<columns;++j) {
     // formula for the ratio is: R = (max/min)^(1/nbBins), with max and min the extrema of the series to bin
     // we need two ratios, one for each series
     // (see Thesis for details)
        double ratio1=0,ratio2=0;
        double min=min_J[j], max=max_J[j];

        // the first series deals with values from 1 to max 
        ratio1=pow(max,1.0/(double)nbBins);
        
        // the first series deals with values from min to 1
        if(ratio2==0) {
           if(min!=0) ratio2=pow(min,1.0/(double)nbBins);
           else ratio2=0.65;
        }
        
        // allocates the array containing the boundaries
        double* boundaries = new double[2*nbBins+1];
        
        // for the part below 1, the boundaries are 1, ratio2, ratio2*ratio2, ..., 0
        boundaries[nbBins]=1;
        for(int k=nbBins-1;k>0;--k) {
           boundaries[k]=boundaries[k+1]*ratio2;
        }
        boundaries[0]=0;
        
        // for the part above 1, the boundaries are 1, ratio1, ratio*ratio1, ..., max
        for(int k=nbBins+1;k<2*nbBins+1;++k) {
           boundaries[k]=boundaries[k-1]*ratio1;
        }
        
        // now that we have the boundaries, we can fill the bins
        int* categorySize = new int[2*nbBins];
        for(int k=0;k<2*nbBins;++k) {
           categorySize[k]=0;
        }
        int cat;
        for(int l=0;l<lines;++l) {
           cat=0;
           while((boundaries[cat+1]<norm(l,j))&&(cat<2*nbBins-1)) {
              ++cat;
           }

           categ(l,j)=cat;
           categorySize[cat]+=1;
        }

        double average=0;
        for(int temp=0;temp<2*nbBins;++temp) average+=categorySize[temp];
        average/=(2*nbBins);
        for(int i=0;i<lines;++i) {
           weights(i,j)=categorySize[(int)(categ(i,j))]-average; // temporary weights
        }
        delete[] boundaries;
        delete[] categorySize;

  } // end of the loop on all conditions

  // now, the weight matrix needs to be "normalised" with costs based on % compared to total size
  double normalizationFactor=lines;
  for(int i=0;i<lines;++i) {
     for(int j=0;j<columns;++j) {
        weights(i,j)=100*weights(i,j)/normalizationFactor;
     }
  }
  
  ofstream outFile;
  string fileNameOut = "weights-"+microarrayFile;
  outFile.open(fileNameOut.c_str(), ios::out);
  if (outFile.bad())
    return (1); // Error while opening... Program stops.

  outFile << lines << " " << columns << endl;
  outFile << weights;

  outFile.close();

  delete[] min_J;
  delete[] max_J;
  delete[] total_J;
  delete[] total_I;

  return 0;
}



/***************************/
/* Dernomalizing a dataset */
/***************************/

int denorm_set(string microarrayFile) {
  ifstream inFile;
  inFile.open(microarrayFile.c_str(), ios::in);

  int lines, columns;

  if (inFile.bad())
    return (1); // Error while opening... Program stops.
  
  inFile >> lines >> columns;
  
  Matrice data(lines,columns,0);  

  for(int i=0;i<lines;++i) {
          for(int j=0;j<columns;++j) {
                  inFile >> data(i,j); // value added to the data matrix
          }
  }
  inFile.close();

  double* normFactor = new double[lines];
  double max;
  for(int i=0;i<lines;++i) {
          max=0;
          normFactor[i]=1;
          for(int j=0;j<columns;++j) {
                  if(max<data(i,j)) max=data(i,j);
          }
          while(max>10*normFactor[i]) normFactor[i]=10*normFactor[i];
  }
  // (normFactor is used to prevent exp(n) with n too large)

  for(int i=0;i<lines;++i) {
          for(int j=0;j<columns;++j) {
                  data(i,j)=exp(data(i,j)/normFactor[i]);
          }
  }
  delete normFactor;
  
  ofstream outFile;
  string fileNameOut;
  cout << "Please enter the name of the file in which to store the denormalized dataset:" << endl;
  cin >> fileNameOut;
  outFile.open(fileNameOut.c_str(), ios::out);

  if (outFile.bad())
    return (1);                                         // Error while opening... Program stops.

  outFile << lines << " " << columns << endl;
  outFile << data;

  outFile.close();

  return 0;
}



/*****************/
/* Main function */
/*****************/

int main(int argc, char *argv[]) {
   int choice, nbBins;
   string microarrayFile;
   int denormNeeded;

   do {
      system("clear");
      cout << "Choose:" << endl;
      cout << "\t1, to create weights using the distribution-based method." << endl;
      cout << "\t2, to \"denormalize\" a dataset (i.e. back from log space)." << endl;
      cout << "\t0, to quit." << endl;
      cin >> choice;
      switch(choice) {
         case 1:
            cout << "Please enter the name of the file containing the microarray data:" << endl;
            cin >> microarrayFile;
            cout << "This must not be applied to datasets containing negative values." << endl;
            cout << "Is denormalization needed? (yes:1 / no: 0)" << endl;
            cin >> denormNeeded;
            if(denormNeeded==0) {
              cout << "Please enter the number of categories for each geometric series:" << endl;
              cin >> nbBins;
              create_weights(microarrayFile,nbBins);
            }
            else {
              cout << "Weight generation cancelled. Please denormalize." << endl;
            }
            break;
         case 2:
            cout << "Please enter the name of the file containing the normalized dataset:" << endl;
            cin >> microarrayFile;
            cout << "This must not be applied to datasets containing positive-only values." << endl;
            cout << "Do you confirm denormalization? (yes:1 / no: 0)" << endl;
            cin >> denormNeeded;
            if(denormNeeded==1)
              denorm_set(microarrayFile);
            else
              cout << "Denormalization cancelled." << endl;
            break;
         default:
            break;
      }
      if(choice) {
         cout << WAIT_MSG << endl;
         int c;
         cin >> c;
      }
   } while(choice);

   return 0;
}


